from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import date, datetime
from typing import Mapping, Sequence

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import Document
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.emails.dropsite_email import DropsiteEmail
from epstein_files.documents.json_file import JsonFile
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile


@dataclass
class DocTypesMixin(ABC):
    """Mixin for classes that maintain a list of `Document`s and want to sift by type."""
    _documents: list[Document] = field(default_factory=list)
    _docs_by_id: dict[str, Document] = field(default_factory=dict)

    @property
    def all_doj_files(self) -> Sequence[DojFile | Email]:
        """All files with the filename EFTAXXXXXX, including those that were turned into `Email` objs."""
        return [d for d in self.documents if d.file_info.is_doj_file]

    @property
    def communications(self) -> Sequence[Communication]:
        """This person's `MessengerLog` and `Email` object."""
        return self.imessage_logs + self.emails

    @property
    def docs_by_id(self) -> Mapping[str, Document]:
        """dict with file IDs as keys and Document objs as values."""
        self._docs_by_id = self._docs_by_id or {doc.file_id: doc for doc in self._documents}
        return self._docs_by_id

    @property
    def documents(self) -> Sequence[Document]:
        return self._documents

    @property
    def doj_files(self) -> list[DojFile]:
        """Only returns DojFile type. Emails derived from DOJ files are not included."""
        return DojFile.filter_for_type(self.other_files)

    @property
    def dropsite_emails(self) -> list[DropsiteEmail]:
        """Older emails from the Dropsite News collection exist as .eml files instead of .txt files."""
        return DropsiteEmail.filter_for_type(self.emails)

    @property
    def earliest_email_at(self) -> datetime:
        return self.emails[0].timestamp

    @property
    def earliest_email_date(self) -> date:
        return self.earliest_email_at.date()

    @property
    def emails(self) -> list[Email]:
        return Email.filter_for_type(self.documents)

    @property
    def imessage_logs(self) -> list[MessengerLog]:
        return MessengerLog.filter_for_type(self.documents)

    @property
    def interesting_other_files(self) -> Sequence[OtherFile]:
        """`OtherFile` objects that have been deemed of interest."""
        return [f for f in self.other_files if f.is_interesting]

    @property
    def json_files(self) -> list[JsonFile]:
        """JSON files from the November document dump, mostly Apple ads related."""
        return JsonFile.filter_for_type(self.other_files)

    @property
    def last_email_at(self) -> datetime:
        return self.emails[-1].timestamp

    @property
    def last_email_date(self) -> date:
        return self.last_email_at.date()

    @property
    def local_extracts(self) -> Sequence[Document]:
        """Returns documents that are locally derived from source files."""
        return [d for d in self.documents if d.file_info.is_local_extract_file]

    @property
    def non_attachments(self) -> Sequence[OtherFile]:
        """Exclude `OtherFile` objs that are attached to `Email` objects."""
        return [d for d in self.other_files if not d._config.attached_to_email_id]

    @property
    def non_json_other_files(self) -> list[OtherFile]:
        return [doc for doc in self.other_files if not isinstance(doc, JsonFile)]

    @property
    def num_docs(self) -> int:
        return len(self.unique_documents)

    @property
    def num_emails(self) -> int:
        return len(self.unique_emails)

    @property
    def other_files(self) -> Sequence[OtherFile]:
        return OtherFile.filter_for_type(self.documents)

    @property
    def unique_documents(self) -> Sequence[Document]:
        """Excludes duplicates and email attachments."""
        return Document.without_dupes(self.documents)

    @property
    def unique_emails(self) -> list[Email]:
        """All `Email` objects except for duplicates."""
        return Document.without_dupes(self.emails)



import json
from collections import defaultdict
from os import unlink
from subprocess import CalledProcessError, check_output
from typing import TypeVar

from rich import box
from rich.padding import Padding
from rich.panel import Panel

from epstein_files.documents.document import Document
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import FIRST_FEW_LINES, OtherFile
from epstein_files.epstein_files import EpsteinFiles
from epstein_files.output.doc_printer import DocPrinter
from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.rich import *
from epstein_files.output.site.internal_links import (AUTHORS_USING_SIGNATURES, FILES_THAT_ARE_NEITHER_EMAILS_NOR,
     HIS_EMAILS, HIS_TEXT_MESSAGES, SELECTIONS_FROM)
from epstein_files.output.site.sites import EMAILERS_TABLE_PNG_PATH, HtmlDir
from epstein_files.output.title_page import print_other_page_link, section_header
from epstein_files.people.interesting_people import EMAILERS_TO_PRINT
from epstein_files.people.names import *
from epstein_files.people.person import Person
from epstein_files.util.constant.html import HTML_TERMINAL_THEME, RICH_HTML_TEMPLATE
from epstein_files.util.constant.strings import AUTHOR
from epstein_files.util.env import args
from epstein_files.util.helpers.data_helpers import dict_sets_to_lists, uniq_sorted
from epstein_files.util.helpers.file_helper import file_size_str, log_file_write
from epstein_files.util.helpers.string_helper import extract_emojis
from epstein_files.util.logging import logger, exit_with_error

DEVICE_SIGNATURE_PADDING = (1, 0)
PRINT_COLOR_KEY_EVERY_N_EMAILS = 150

OTHER_INTERESTING_EMAILS_SUBTITLE = 'Other Interesting Emails\n(these emails have been flagged as being of particular interest)'
CONVERSATIONS_SORTED_BY_TXT = Text("(conversations sorted chronologically based on timestamp of the first text message)", 'dim')
DEVICE_SIGNATURE_SUBTITLE = f"Email [italic]Sent from \\[DEVICE][/italic] Signature Breakdown"
DEVICE_SIGNATURE = 'Device Signature'

T = TypeVar('T')


def print_all_emails_chronological(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """Print all non-mailing list emails in chronological order."""
    emails = Document.sort_by_timestamp([e for e in epstein_files.unique_emails if not e.is_mailing_list])
    emails = _max_records(emails)
    title = f'Table of All {len(emails):,} Non-Junk Emails in Chronological Order (actual emails below)'
    table = Email.build_emails_table(emails, title=title, show_length=True)
    printer.print(Padding(table, (2, 0)))
    printer.print_section_subtitle('The Chronologically Ordered Emails')
    printer.print_documents(emails)


def print_curated_chronological(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """Print only interesting files of all types in chronological order."""
    logger.warning(f'Printing curated chronological site...')

    def should_print(doc: Document) -> bool:
        if doc._config.attached_to_email_id:
            return False
        else:
            return bool(doc.is_interesting if not args.invert_chrono else not doc.is_interesting)

    docs = [d for d in epstein_files.unique_documents if should_print(d)]
    printer.print_section_subtitle('Selected Files of Interest in Chronological Order')
    printer.print_documents(_max_records(docs))


def print_doj_files(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """Doesn't print DojFiles that are actually Emails, that's handled in print_emails()."""
    printer.collect_other_files_to_tables = False
    docs = Document.without_dupes(epstein_files.doj_files)
    printer.print_documents(_max_records(docs))


def print_emailers_info(epstein_files: EpsteinFiles) -> None:
    """Print tbe summary table of everyone who sent or received an email to a .png file."""
    printer = DocPrinter(epstein_files=epstein_files)
    printer.print_color_key()
    console.line()
    all_emailers = sorted(epstein_files.emailers, key=lambda person: person.sort_key)
    console.print(Person.emailer_info_table(all_emailers, show_epstein_total=True))

    if not args.build:
        logger.warning(f"Not writing .png file because --build is not set")
        return

    svg_path = f"{EMAILERS_TABLE_PNG_PATH}.svg"
    console.save_svg(svg_path, theme=HTML_TERMINAL_THEME, title="Epstein Emailers")
    log_file_write(svg_path)

    try:  # Inkscape is better at converting svg to png
        inkscape_cmd_args = ['inkscape', f'--export-filename={EMAILERS_TABLE_PNG_PATH}', svg_path]
        logger.warning(f"Running inkscape cmd: {' '.join(inkscape_cmd_args)}")
        check_output(inkscape_cmd_args)
    except (CalledProcessError, FileNotFoundError) as e:
        logger.error(f"Failed to convert svg to png with inkscape, falling back to cairosvg: {e}")
        import cairosvg
        cairosvg.svg2png(url=svg_path, write_to=str(EMAILERS_TABLE_PNG_PATH))

    log_file_write(EMAILERS_TABLE_PNG_PATH)
    unlink(svg_path)


def print_emails_section(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """
    Print emails grouped by participant with summary tables.
    Emails can be printed more than once (e.g. in the sender's section and each recipient's).
    """
    printer.print(section_header((SELECTIONS_FROM if not args.all_emails else '') + HIS_EMAILS))
    all_emailers = sorted(epstein_files.emailers, key=lambda person: person.earliest_email_at)
    num_since_color_key = 0

    if args.names:
        try:
            emailers_to_print = epstein_files.person_objs(args.names)
        except Exception as e:
            exit_with_error(str(e))
    else:
        emailers_to_print = all_emailers if args.all_emails else epstein_files.person_objs(EMAILERS_TO_PRINT)
        printer.print(_section_summary_table(Person.emailer_info_table(all_emailers, emailers_to_print)))

    for person in _max_records(emailers_to_print):
        if person.is_interesting is False and not args.names:
            logger.warning(f"Skipping person with is_interesting=False '{person.name}'")
            continue

        printed_emails = person.print_emails(printer)

        # Print color key every once in a while
        if (num_since_color_key := num_since_color_key + len(printed_emails)) > PRINT_COLOR_KEY_EVERY_N_EMAILS:
            printer.print_color_key()
            num_since_color_key = 0

    if args.names:
        return

    # Print other interesting emails
    extra_emails = [
        e for e in epstein_files.unique_emails
        if e.is_interesting and e.file_id not in printer.printed_ids
    ]

    if len(extra_emails) > 0:
        logger.warning(f"Found {len(extra_emails)} additional interesting emails by less interesting people...")
        printer.print_section_subtitle(OTHER_INTERESTING_EMAILS_SUBTITLE)
        printer.print_documents(Document.sort_by_timestamp(extra_emails))

    print_signatures_and_emojis(epstein_files, printer)
    fwded_articles = [e for e in printer.printed_emails if e.is_fwded_article]
    log_msg = f"Rewrote {len(Email.rewritten_header_ids)} of {len(printer.printed_emails)} email headers"
    logger.warning(f"  -> {log_msg}, {len(fwded_articles)} of the Emails printed were forwarded articles.")

    if args.all_emails:
        _verify_all_emails_were_printed(epstein_files, printer.printed_emails)


# NOTE: the JSON files from Nov. 2025 are completely uninteresting
def print_json_files(epstein_files: EpsteinFiles) -> None:
    """Print all the `JsonFile` objects to a unified JSON file."""
    if args.build:
        json_data = {jf.file_info.url_slug: jf.json_data() for jf in epstein_files.json_files}
        # TODO: stopgap so this doesn't break but we're no longer building this file.
        build_path = 'json_files.json' # SiteType.build_path('json_files')# SiteType.JSON_FILES)

        with open(build_path, 'wt') as f:
            f.write(json.dumps(json_data, sort_keys=True))
            log_file_write(build_path)
    else:
        for json_file in epstein_files.json_files:
            console.line(2)
            console.print(json_file._summary_panel)
            console.print_json(json_file.json_str(), indent=4, sort_keys=False)


def print_json_metadata(epstein_files: EpsteinFiles) -> None:
    """Print all our `DocCfg` and derived info about authorship etc."""
    if args.build:
        output_path = Site.html_output_path(Site.JSON_METADATA)

        with open(output_path, 'wt') as f:
            f.write(epstein_files.json_metadata())
            log_file_write(output_path)
    else:
        console.print_json(epstein_files.json_metadata(), indent=4, sort_keys=True)


def print_other_files_section(epstein_files: EpsteinFiles, printer: DocPrinter) -> list[OtherFile]:
    """Prints a table of `OtherFile` and returns the objects that were printed."""
    if args.uninteresting:
        files = [f for f in epstein_files.non_attachments if not f.is_interesting]
    else:
        files = [f for f in epstein_files.non_attachments if args.all_other_files or f.is_interesting]

    files = _max_records(Document.sort_by_timestamp(files))
    title_pfx = '' if args.all_other_files else 'Selected '
    category_table = OtherFile.summary_table(files, title_pfx=title_pfx)
    printer.print_section_subtitle(f"{FIRST_FEW_LINES} of {len(files)} {title_pfx}{FILES_THAT_ARE_NEITHER_EMAILS_NOR}")
    print_other_page_link(epstein_files)  # TODO: not in custom HTML
    printer.print(_section_summary_table(category_table))

    # If --all-other-files is enables, print the biographical panels, otherwise just print a big table
    if args.all_other_files:
        printer.print_documents(files, suppressed_as_normal=True)
    else:
        printer.print_centered(OtherFile.files_preview_table(files, title_pfx=title_pfx))

    return files


def print_signatures_and_emojis(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """Print device signatures and emojis and who used them."""
    printer.print_section_subtitle(DEVICE_SIGNATURE_SUBTITLE)
    printer.print(_signature_table(epstein_files.email_device_signatures_to_authors(), (DEVICE_SIGNATURE, AUTHOR), ', '))
    printer.print(_signature_table(epstein_files.email_authors_to_device_signatures(), (AUTHOR, DEVICE_SIGNATURE)))
    author_emojis = defaultdict(set)
    emoji_authors = defaultdict(set)

    for email in [e for e in epstein_files.emails if e.is_word_count_worthy]:
        if (emojis := extract_emojis(email.actual_text)):
            author_emojis[email.author].update(emojis)

            for emoji in emojis:
                emoji_authors[emoji].add(email.author_str)

    printer.line()
    printer.print_section_subtitle("Emoji Usage")
    printer.print_centered(_signature_table(emoji_authors, ('Emoji', AUTHOR)))


def print_stats(epstein_files: EpsteinFiles) -> None:
    """Used to generate fixture data for `pytest`."""
    console.print('\n\n\n', Panel('JSON Stats Dump', expand=True, style='reverse bold'), '\n')
    print_json(MessengerLog.count_authors(epstein_files.imessage_logs), f"MessengerLog Sender Counts", skip_falsey=True)
    print_json(epstein_files.email_author_counts(), f"Email Author Counts", skip_falsey=True)
    print_json(epstein_files.email_recipient_counts(), f"Email Recipient Counts", skip_falsey=True)
    print_json(epstein_files.email_signature_substitution_counts(), "Email signature_substitution_countss", skip_falsey=True)
    print_json(dict_sets_to_lists(epstein_files.email_authors_to_device_signatures()), "Email author device signatures")
    print_json(dict_sets_to_lists(epstein_files.email_device_signatures_to_authors()), "Email authors by device")
    unknown_recipient_ids = sorted([e.file_id for e in epstein_files.emails if None in e.recipients or not e.recipients])
    print_json(unknown_recipient_ids, "Unknown Recipient IDs")
    print_json(Document.count_by_month(epstein_files.documents), "All documents count by month")
    print_json(sorted([f.file_id for f in epstein_files.interesting_other_files]), "Interesting OtherFile IDs")
    print_json(highlighter.highlight_counts, f"Highlight Counts")
    print_json(epstein_files.counterparties_dict, f"Counterparties")
    highlighter.print_highlight_counts(console)


def print_text_msgs_section(epstein_files: EpsteinFiles, printer: DocPrinter) -> None:
    """Print `MessengerLog` objects and return objects that were printed."""
    section_header_panel = section_header((SELECTIONS_FROM if not args.all_texts else '') + HIS_TEXT_MESSAGES)
    printer.print(section_header_panel)
    printer.print(Align.center(CONVERSATIONS_SORTED_BY_TXT))

    if args.names:
        imessage_logs = [log for log in epstein_files.imessage_logs if log.author in args.names]
    else:
        imessage_logs = [log for log in epstein_files.imessage_logs if args.all_texts or log.is_interesting]

    imessage_logs = _max_records(imessage_logs)

    if not imessage_logs:
        logger.warning(f"No MessengerLog found for {args.names}")
        return

    if not args.names:
        printer.print(_section_summary_table(MessengerLog.summary_table(imessage_logs)))

    printer.print_documents(imessage_logs)


def show_urls() -> None:
    """Print the various URLs generated by this code."""
    try:
        urls = {
            k: Text('[', 'grey30').append(file_size_str(Site.html_output_path(k), 1), 'cyan').append('] ') + \
                Text(v, ARCHIVE_LINK_COLOR)
            for k, v in Site.all_urls().items()
        }
    except FileNotFoundError:
        urls = Site.all_urls()

    console.print(Padding(styled_dict(urls), (1)))


def write_html(output_path: Path | Site | None, **kwargs) -> Path | None:
    """
    Write all `console` output to HTML in `output_path` (if provided).
    If `args.write_txt` is set colored ANSI `.txt` files will be written instead.
    Returns the path that was written (if any).
    """
    if not output_path:
        logger.warning(f"Not writing HTML (args.build={args.build}, args._site={args._site}.")
        return
    elif isinstance(output_path, Site):
        output_path = Site.html_output_path(output_path)

    if args.write_txt:
        output_path = HtmlDir.build_path(f"{output_path.stem}.txt")
        console.save_text(str(output_path))
    else:
        console.save_html(
            str(output_path),
            clear=False,
            code_format=str(RICH_HTML_TEMPLATE),
            theme=HTML_TERMINAL_THEME,
            **kwargs
        )

    log_file_write(output_path)
    return output_path


def _section_summary_table(table: Table) -> Align:
    return Align(Padding(table, (1, 0, 1, 0)), 'center')


def _signature_table(keyed_sets: dict[str, set[str]], cols: tuple[str, str], join_char: str = '\n') -> Padding:
    """Build table for who signed emails with 'Sent from my iPhone' etc."""
    new_dict = dict_sets_to_lists(keyed_sets)

    if cols[0] == AUTHOR:
        if cols[1] == DEVICE_SIGNATURE:
            title = 'Email Signatures Used By Authors'
        else:
            title = f"{cols[1]} used by Authors"
    else:
        title = AUTHORS_USING_SIGNATURES

    table = build_table(title, header_style="bold reverse", show_lines=True)

    for i, col in enumerate(cols):
        table.add_column(col.title() + ('s' if i == 1 else ''))

    for k in sorted(new_dict.keys()):
        table.add_row(highlighter(k or UNKNOWN), highlighter(join_char.join(sorted(new_dict[k]))))

    return Padding(table, DEVICE_SIGNATURE_PADDING)


def _verify_all_emails_were_printed(epstein_files: EpsteinFiles, printed_emails: list[Email]) -> None:
    """Log warnings if some emails were never printed."""
    email_ids_that_were_printed = set([email.file_id for email in printed_emails])
    logger.warning(f"Printed {len(printed_emails):,} emails of {len(email_ids_that_were_printed):,} unique file IDs.")
    missed_an_email = False

    for email in epstein_files.unique_emails:
        if email.file_id not in email_ids_that_were_printed:
            logger.error(f"Failed to print {email._summary}")
            missed_an_email = True

    if not missed_an_email:
        logger.warning(f"All {len(epstein_files.emails):,} emails printed at least once.")


def _max_records(docs: list[T]) -> list[T]:
    """Truncate number of Documents if `args.max_records` is specified."""
    if args.max_records and args.max_records < len(docs):
        logger.warning(f"Truncating to args.max_records={args.max_records} objects...")
        return docs[0:args.max_records]
    else:
        return docs



"""
Constants and methods for identifying people in email headers.
"""
import re
from collections import Counter
from typing import Optional

from epstein_files.documents.documents.categories import is_uninteresting
from epstein_files.documents.emails.constants import UNINTERESTING_EMAILERS
from epstein_files.output.highlight_config import HIGHLIGHTED_ENTITIES
from epstein_files.people.entity import Entity, Organization
from epstein_files.people.names import *
from epstein_files.util.constant.strings import REDACTED
from epstein_files.util.helpers.data_helpers import escape_single_quotes, flatten, groupby, uniq_sorted, without_falsey
from epstein_files.util.helpers.string_helper import as_pattern
from epstein_files.util.logging import logger

BAD_EMAILER_REGEX = re.compile(r'^(>|11111111)|agreed|ok(?!asha)|sexy|re:|fwd:|LIMITED PARTNERS|Multiple Senders|((sent|attachments|subject|importance).*|.*(january|201\d|hysterical|i have|image0|so that people|article 1.?|PROSPECTIVE INVESTORS|momminnemummin|These conspiracy theories|your state|undisclosed|www\.theguardian|talk in|it was a|what do|cc:|call (back|me)|afiaata|[IM]{4,}).*)$', re.IGNORECASE)
BAD_NAME_CHARS_REGEX = re.compile(r"[\"'\[\]*><•=()‹?]")
TIME_REGEX = re.compile(r'^((\d{1,2}/\d{1,2}/\d{2,4}|Thursday|Monday|Tuesday|Wednesday|Friday|Saturday|Sunday)|\d{4} ).*')

# Unhighlighted / uncategorized emailers we don't know much about but need regexes to identify
ADDITIONAL_EMAILERS = [
    # Custom regex
    Entity('BS Stern', emailer_pattern=r"BS Ste(m|rn)"),
    Entity(INTELLIGENCE_SQUARED, emailer_pattern=r"intelligence\s*squared"),
    Entity('Matthew Schafer', emailer_pattern=r"matthew\.?schafer?"),
    Entity(MICHAEL_BUCHHOLTZ, emailer_pattern=r"Michael.*Buchholtz"),
    Entity(SAMUEL_LEFF, emailer_pattern=r"Sam(uel)?(/Walli)? Leff"),
    Entity(THANU_BOONYAWATANA, emailer_pattern=r"Thanu (BOONYAWATANA|Cnx)"),
    # No custom regex
    Entity('Anne Boyles', match_partial=None),
    Entity('Ariane Dwyer'),
    Entity('Brittany Henderson'),
    Entity('Danny Goldberg', match_partial=None),
    Entity('Jeff Pagliuca'),
    Entity(JOHN_PAGE, match_partial=None),
    Entity('Julie Shample'),
    Entity('Kathleen Ruderman'),
    Entity('Kevin Bright', match_partial=None),
    Entity('Larry Cohen', match_partial=None),
    Entity('Lawrence Delson'),
    Entity('Martin Zeman'),
    Entity('Michael Simmons', match_partial=None),
    Entity('middle.east.update@hotmail.com'),
    Entity('Nancy Cain'),
    Entity('Nancy Portland', match_partial=None),
    Entity(OLIVER_GOODENOUGH),
    Entity('Peter Green', match_partial=None),
    Entity('Sarah Mapes'),
    # Krassner CCs
    Entity('Barb Cowles', match_partial=None, is_interesting=False),
    Entity('Bob Fass', match_partial=None, is_interesting=False),
    Entity('Caryl Ratner', match_partial=None, is_interesting=False),
    Entity('Daniel Dawson', match_partial=None, is_interesting=False),
    Entity('Holly Krassner Dawson', match_partial=None, is_interesting=False),
    Entity('Lee Quarnstrom', match_partial=None, is_interesting=False),
    Entity('Linda W. Grossman', match_partial=None, is_interesting=False),
    Entity('Lynnie Tofte Fass', match_partial=None, is_interesting=False),
    Entity('Marie Moneysmith', match_partial=None, is_interesting=False),
    Entity('Tom', match_partial=None, is_interesting=False, is_emailer=False),
    Entity('Harry Shearer', match_partial=None, is_interesting=False),
    Entity('Jay Levin', match_partial=None, is_interesting=False),
    Entity('Lanny Swerdlow', match_partial=None, is_interesting=False),
    Entity('Larry Sloman', match_partial=None, is_interesting=False),
    Entity('W&K', match_partial=None, is_interesting=False),
    Entity('Walli Leff', match_partial=None, is_interesting=False),
    Entity('Mrisman02', match_partial=None, is_interesting=False),
    Entity('Nick Kazan', match_partial=None, is_interesting=False),
    Entity('Rebecca Risman', match_partial=None, is_interesting=False),
]

SUPPRESS_LOGS_FOR_AUTHORS = [
    'Multiple Senders Multiple Senders',
    'Undisclosed recipients:',
    'undisclosed-recipients:',
]

# Collect all configured entities into various data structures
CONFIGURED_ENTITIES = HIGHLIGHTED_ENTITIES + ADDITIONAL_EMAILERS
ENTITIES_DICT = {c.name: c for c in CONFIGURED_ENTITIES}

for name in UNINTERESTING_EMAILERS:
    if (entity := ENTITIES_DICT.get(name)):
        entity._debug_log(f"Found UNINTERESTING_EMAILER, setting is_interesting=False...")  # TODO: doesn't mean much right now
        entity.is_interesting = False
    else:
        CONFIGURED_ENTITIES.append(Entity(name, is_interesting=False, match_partial=None))
        CONFIGURED_ENTITIES[-1]._debug_log(f"Created new Entity for UNINTERESTING_EMAILER entry...")

ENTITIES_DICT = {c.name: c for c in CONFIGURED_ENTITIES}  # Rebuild with any new uninteresting mailers
EMAILER_REGEXES = {c.name: c.emailer_regex for c in CONFIGURED_ENTITIES if c.is_emailer}
ENTITY_CATEGORIES = groupby(CONFIGURED_ENTITIES, lambda entity: entity.category)

# TODO: dict of names that are configured but have no Entity. This is filled in in epstein_files.py which sucks.
CONFIGURED_NON_ENTITIES: dict[str, Entity] = {}
UNCONFIGURED_ENTITIES_ENCOUNTERED: dict[str, Entity] = {}

if len(CONFIGURED_ENTITIES) != len(ENTITIES_DICT):
    counts = Counter([c.name for c in CONFIGURED_ENTITIES])
    more_than_one = [k for k, v in counts.items() if v > 1]
    raise ValueError(f"{len(CONFIGURED_ENTITIES)} entities but only {len(ENTITIES_DICT)} names! Bad names: {more_than_one}")


# Strings that usually signify an identity if present in email body
IDENTIFYING_REGEXES = {}

for entity in CONFIGURED_ENTITIES:
    for identifier in entity.identifying_strings:
        IDENTIFYING_REGEXES[re.compile(as_pattern(identifier), re.IGNORECASE)] = entity

# file IDs that contain a unique signifier but do not involve that person
IDENTIFIER_FALSE_ALARMS = ['EFTA00961792']

# Elements of this list will not trigger warnings in get_entity when they are found missing
NO_WARNING_NAMES = [
    '',
    'American Express Travel',
    # 'Leo',
    'Unik',
    'Vlad',
    'karen',
]


def cleanup_str(s: str) -> str:
    return BAD_NAME_CHARS_REGEX.sub('', s.replace(REDACTED, '')).strip().strip('_').strip()


def extract_emailer_names(emailer_str: str) -> list[Name]:
    """Return a list of people's names found in `emailer_str` (email author or recipients field)."""
    raw_names = emailer_str.split(';')
    emailer_str = cleanup_str(emailer_str)

    if len(emailer_str) == 0:
        return []
    elif raw_names == [REDACTED] or raw_names == [UNKNOWN]:
        return [None]
    elif emailer_str.lower() in ['j', 'jeff', 'jeffrey']:
        return [JEFFREY_EPSTEIN]
    elif emailer_str.lower() in ['sa', 's a']:
        return [SHAHER_ABDULHAK_BESHER]

    for name in EMAILER_REGEXES.keys():
        if name == 'Tom':
            print(f"\n\n Tom: {EMAILER_REGEXES[name]}\n\n")

    names_found: list[Name] = [name for name, regex in EMAILER_REGEXES.items() if regex.search(emailer_str)]

    if len(emailer_str) <= 2 or BAD_EMAILER_REGEX.match(emailer_str) or TIME_REGEX.match(emailer_str):
        if len(names_found) == 0 and emailer_str not in SUPPRESS_LOGS_FOR_AUTHORS:
            logger.warning(f"No emailer found in '{escape_single_quotes(emailer_str)}'")
        else:
            logger.info(f"Extracted {len(names_found)} names from semi-invalid '{emailer_str}': {names_found}...")

        return names_found

    names_found = [reverse_first_and_last_names(name) for name in (names_found or [emailer_str])]
    logger.debug(f"names_found in '{emailer_str}': {names_found}")
    return names_found


def get_entity(name: str | Entity, doc: Optional['Document'] = None) -> Entity:
    if isinstance(name, Entity):
        return name
    elif name in ENTITIES_DICT:
        return ENTITIES_DICT[name]
    elif name in CONFIGURED_NON_ENTITIES:
        return CONFIGURED_NON_ENTITIES[name]  # Avoids spurious warnings
    elif name not in UNCONFIGURED_ENTITIES_ENCOUNTERED:
        if name not in NO_WARNING_NAMES and '@' not in name:
            log = doc._warn if doc else logger.warning
            log(f"Encountered unconfigured entity name '{name}'")

        UNCONFIGURED_ENTITIES_ENCOUNTERED[name] = Entity(name)

    return UNCONFIGURED_ENTITIES_ENCOUNTERED[name]


def get_entities(_names: Sequence[Name | Entity], doc: Optional['Document'] = None) -> list[Entity]:
    """Also uniquifies and removes None / empty string."""
    names = Entity.coerce_entity_names(without_falsey(_names))
    return [get_entity(name, doc) for name in uniq_sorted(names)]



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from epstein_files.util.logging import logger


class LoggingEntity(ABC):
    """
    Classes that implement `_identifier()` or overload `_log_prefix()` can call self._log(), self._warn(), etc.
    """

    @property
    def _class_name(self) -> str:
        return type(self).__name__

    @property
    @abstractmethod
    def _identifier(self) -> str:
        """Overload for a reusable approach to  _log_prefix, e.g. Document(12345) or Person('James')"""
        pass

    @property
    def _log_prefix(self) -> str:
        return f"{self._class_name}({self._identifier})"

    def _debug_log(self, msg: str) -> None:
        self._log(msg, logging.DEBUG)

    def _log(self, msg: str, level: int = logging.INFO) -> None:
        logger.log(level, f"{self._log_prefix} {msg}")

    def _warn(self, msg: str) -> None:
        self._log(msg, logging.WARNING)



"""
HTML file paths and URLs for files built by `epstein_generate`.
"""
import shutil
from enum import auto, StrEnum
from pathlib import Path
from typing import Self

from rich.markup import escape
from rich.text import Text

from epstein_files.util.constant.strings import (AUX_SITE_LINK_STYLE, CHRONOLOGICAL, DOJ_2026_TRANCHE,
     EPSTEIN_FILES_NOV_2025, HOUSE_OVERSIGHT_TRANCHE)
from epstein_files.util.external_link import link_text_obj, parenthesize
from epstein_files.util.logging import logger

DEFAULT_HTML_DIR = Path('docs')
EMAILERS_TABLE_PNG_PATH = DEFAULT_HTML_DIR.joinpath('emailers_info_table.png')

# Github URLs
GH_REPO_NAME = 'epstein_text_messages'
GH_PAGES_BASE_URL = 'https://michelcrypt4d4mus.github.io'
GH_PROJECT_URL = f'https://github.com/michelcrypt4d4mus/{GH_REPO_NAME}'
GH_MASTER_URL = f"{GH_PROJECT_URL}/blob/master"
ATTRIBUTIONS_URL = f'{GH_MASTER_URL}/epstein_files/util/constants.py'
EXTRACTS_BASE_URL = f'{GH_MASTER_URL}/emails_extracted_from_legal_filings'
BASE_URL = f"{GH_PAGES_BASE_URL}/{GH_REPO_NAME}"

CUSTOM_HTML_PREFIX = 'real_html_'
NAMES_PREFIX = 'only_names_'
MOBILE_SUFFIX = '_mobile'


class Site(StrEnum):
    BIOGRAPHIES = auto()
    COLORS_ONLY = auto()
    CURATED = auto()
    CURATED_MOBILE = auto()
    CHRONOLOGICAL = CHRONOLOGICAL
    CHRONOLOGICAL_MOBILE = f"{CHRONOLOGICAL}_mobile"
    DEVICE_SIGNATURES = auto()
    DOJ_FILES = auto()
    EMAILS = auto()
    EMAILS_CHRONOLOGICAL = auto()
    JSON_METADATA = auto()
    NAMES = auto()
    OTHER_FILES_TABLE = auto()
    SAMPLE = auto()
    TEXT_MESSAGES = auto()
    WORD_COUNT = auto()
    # Dev sites
    DEV_SAMPLE = auto()

    @classmethod
    def all_urls(cls) -> dict[Self, str]:
        return {site: cls.get_url(site) for site in cls}

    @classmethod
    def all_links(cls) -> dict['Site', Text]:
        """Use `SITE_DESCRIPTIONS` dict because it's ordered and also omits unpublished site types."""
        return {site: Site.link_txt(site) for site in SITE_DESCRIPTIONS}

    @classmethod
    def custom_html_build_path(cls, site: 'Site | Path') -> Path:
        """Path for the templated custom HTML version of the page."""
        filename = site.name if isinstance(site, Path) else cls.html_output_path(site).name
        return HtmlDir.build_path(f'{CUSTOM_HTML_PREFIX}{filename}')

    @classmethod
    def directory_links(cls) -> dict['Site', Text]:
        """Everything but mobile."""
        return {k: v for k, v in cls.all_links().items() if not cls.is_mobile(k)}

    @classmethod
    def html_output_path(cls, _site: 'Site') -> Path:
        """Defaults to `[site].html` if not configured in `HTML_BUILD_FILENAMES`."""
        if _site == cls.NAMES:
            from epstein_files.util.env import args
            names = sorted(['unknown' if n is None else n for n in args.names])
            site = NAMES_PREFIX + '__'.join(sorted(names)).replace(' ', '_').lower()
        else:
            site = _site

        return HtmlDir.build_path(HTML_BUILD_FILENAMES.get(site, f"{site}.html"))

    @classmethod
    def html_output_path_mobile(cls, site: 'Site') -> Path:
        output_path = cls.html_output_path(site)
        return output_path.parent.joinpath(output_path.name.replace('.html', '_mobile.html'))

    @classmethod
    def get_url(cls, site: 'Site') -> str:
        return f"{BASE_URL}/{cls.html_output_path(site).name}"

    @classmethod
    def get_mobile_redirect_url(cls, site: Self) -> str:
        """Mobile defaults to chronological."""
        if cls.is_mobile(site) or site == cls.NAMES:
            return cls.get_url(site)
        elif site == cls.CURATED:
            redirect_type = cls.CURATED_MOBILE
        else:
            redirect_type = cls.CHRONOLOGICAL_MOBILE  # TODO: the curated site is breaking iPhone simulator so no redirect for now

        return cls.get_url(redirect_type)

    @classmethod
    def is_chronoligical(cls, site: Self) -> bool:
        return CHRONOLOGICAL in site

    @classmethod
    def is_mobile(cls, site: Self) -> bool:
        return site.endswith(MOBILE_SUFFIX)

    @classmethod
    def link_txt(cls, site: Self) -> Text:
        description = SITE_DESCRIPTIONS[site]
        extra_info = ''

        if cls._is_lesser_site(site):
            style = 'gray30'# 'light_pink4'
        else:
            style = AUX_SITE_LINK_STYLE

        if ':' in description:
            description, extra_info = SITE_DESCRIPTIONS[site].split(':')
            extra_info = Text(escape(extra_info), style=f'plum4 italic')
            extra_info = Text(' ').append(parenthesize(extra_info, 'color(147) dim'))

        style_mod = '' if cls._is_lesser_site(site) else 'bold'
        link = link_text_obj(Site.get_url(site), escape(description), f"{style} {style_mod}")
        return link.append(extra_info)

    @classmethod
    def mobile_compatible_types(cls) -> list['Site']:
        """Site types that have a valid mobile equivalent."""
        mobile_sites = [site for site in cls if cls.is_mobile(site)]
        return [site.removesuffix(MOBILE_SUFFIX) for site in mobile_sites]

    @classmethod
    def _is_lesser_site(cls, site: Self) -> bool:
        return site in [cls.DOJ_FILES, cls.JSON_METADATA]  #+ [cls.JSON_FILES]


class HtmlDir:
    """Container to hold state of HTML_DIR value because we can't add class vars to `Site` enum."""
    HTML_DIR = DEFAULT_HTML_DIR

    @classmethod
    def build_path(cls, filename: str) -> Path:
        return cls.HTML_DIR.joinpath(filename)


# TODO: purge repo of old files for space:
#  - curated_chronological.html
HTML_BUILD_FILENAMES = {
    Site.EMAILS_CHRONOLOGICAL:  f'chronological_emails.html',
    Site.CHRONOLOGICAL:         f"index.html",
    Site.CHRONOLOGICAL_MOBILE:  f"mobile_chronological.html",
    Site.DOJ_FILES:             f'doj_2026-01-30_non_email_files.html',
    Site.EMAILS:                f'emails_grouped_by_counterparty.html',
    Site.JSON_METADATA:         f'metadata.json',
    Site.TEXT_MESSAGES:         f'text_messages_{EPSTEIN_FILES_NOV_2025}.html',
    Site.WORD_COUNT:            f'communication_word_count.html',
#     SiteType.EPSTEIN_WORD_COUNT: 'epstein_texts_and_emails_word_count.html'),
}

# NOTE: Order matters, it's the order the links are shown in the header
# Colons are used to break and parenthesize display
SITE_DESCRIPTIONS = {
    Site.CHRONOLOGICAL:         f"chronological curated:all types intermingled",
    Site.CURATED:               f"emailers curated:emails grouped by person of interest",
    Site.EMAILS:                f"emailers:all emails grouped by person",
    Site.EMAILS_CHRONOLOGICAL:  f"emails chronological:all emails chronological order",
    Site.BIOGRAPHIES:           f"people:one line biographies with some links",
    Site.DEVICE_SIGNATURES:     f"signatures:email signatures/emojis and who uses them",
    Site.TEXT_MESSAGES:         f"text messages:{HOUSE_OVERSIGHT_TRANCHE}",
    Site.CURATED_MOBILE:        f"mobile:an attempt at mobile compatibility",
    Site.CHRONOLOGICAL_MOBILE:  f"chrono mobile:another attempt at mobile compatibility",
    Site.OTHER_FILES_TABLE:     f"other:files that are not emails or texts",
    Site.WORD_COUNT:            f"word count:of Epstein's communications",
    Site.DOJ_FILES:             f"doj files:raw OCR text {DOJ_2026_TRANCHE}",
    Site.JSON_METADATA:         f"metadata:author bios, attribution explanations",
}

# These are site types where the custom HTML is good enough to deploy
DEPLOY_CUSTOM_HTML_SITES = [
    Site.BIOGRAPHIES,
    Site.CHRONOLOGICAL,
    Site.CHRONOLOGICAL_MOBILE,
    Site.EMAILS_CHRONOLOGICAL,
    Site.OTHER_FILES_TABLE,
]


def make_clean() -> None:
    """Delete all build artifacts."""
    for site in Site:
        for build_file in [Site.html_output_path(site), Site.custom_html_build_path(site)]:
            if site == Site.NAMES:
                paths = [f for f in build_file.parent.glob(f"{build_file.stem}*.html")]
            else:
                paths = [build_file, Path(f"{build_file}.txt")]

            for file in [f for f in paths if f.exists()]:
                logger.warning(f"Removing build file '{file}'...")
                file.unlink()


def use_custom_html() -> None:
    """Overwrite normal rich html export with custom HTML for all `DEPLOY_CUSTOM_HTML_SITES`."""
    for site in DEPLOY_CUSTOM_HTML_SITES:
        from_path = Site.custom_html_build_path(site)
        to_path = Site.html_output_path(site)

        if from_path.exists():
            logger.warning(f"Copying/overwriting '{from_path}' to '{to_path}'...")
            shutil.copy2(from_path, to_path)
        else:
            logger.error(f"No custom HTML file found at '{from_path}'")



import json
from copy import deepcopy
from dataclasses import asdict, dataclass, field, fields
from datetime import datetime
from pathlib import Path
from typing import Any, Generator, Literal, Self, Sequence

from dateutil import tz
from dateutil.parser import parse
from rich.text import Text

from epstein_files.documents.documents.categories import (Category, Interesting, Neutral, Uninteresting,
     is_category, is_interesting, is_uninteresting)
from epstein_files.output.site.site_config import MobileConfig
from epstein_files.people.interesting_people import PERSONS_OF_INTEREST
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import CharRange, coerce_utc_strict, without_falsey
from epstein_files.util.helpers.file_helper import is_doj_file
from epstein_files.util.external_link import ExternalLink
from epstein_files.util.helpers.string_helper import collapse_whitespace, is_bool_prop, join_truthy, quote
from epstein_files.util.logging import logger
from epstein_files.util.logging_entity import LoggingEntity

DebugDict = dict[str, bool | datetime | set | str | Path | None]
DuplicateType = Literal['bounced', 'earlier', 'quoted', 'redacted', 'same']
Metadata = dict[str, bool | datetime | int | str | None | list[str | None] | dict[str, bool | str]]

DEFAULT_TRUNCATE_TO = 4000
SHORT_TRUNCATE_TO = int(DEFAULT_TRUNCATE_TO / 3)
NO_TRUNCATE = -1
MAX_REPR_LINE_LENGTH = 135

CHECK_LINK_FOR_DETAILS = 'not shown here, check original PDF for details'
QUOTE_PREFIX = 'see quote'
SAME = 'same'

ZUBAIR_AND_ANYA = f"{ZUBAIR_KHAN} and Anya Rasulova"

# Authors of financial report pablum
FINANCIAL_REPORTS_AUTHORS = [
    BOFA_MERRILL,
    DEUTSCHE_BANK,
    ELECTRON_CAPITAL_PARTNERS,
    GOLDMAN_INVESTMENT_MGMT,
    'Invesco',
    JP_MORGAN,
    'Morgan Stanley',
    'S&P',
]

# Description prefixes we are uninterested in
UNINTERESTING_PREFIXES = [
    'article about',
    CVRA,
    f"{HARVARD} Econ",
    HARVARD_POETRY,
    JASTA,
    LEXIS_NEXIS,
    PALM_BEACH_TSV,
    'US Office',
]

DUPE_TYPE_STRS: dict[DuplicateType, str] = {
    'bounced': 'a bounced copy of',
    'earlier': 'an earlier draft of',
    'quoted': 'quoted in full in',
    'redacted': 'a redacted version of',
    SAME: 'the same as',
}

# only used to order fields in metadtaa and repr()
FIELD_SORT_KEY = {
    'id': 'a',
    AUTHOR: 'aa',
    'comment': 'zz',
    'duplicate_ids': 'dup',
    'duplicate_of_id': 'dupe',
    'recipients': 'aaa',
}

# Fields like timestamp and author are better added from the Document object
NON_METADATA_FIELDS = [
    'actual_text',
    'id',
    'is_synthetic',
    'display_text',
]

# Categories where we want to include the category name at start of the description string
CATEGORY_PREAMBLES = {
    Interesting.DIARY: VICTIM_DIARY,
    Interesting.REPUTATION: REPUTATION_MGMT,
    Neutral.DEPOSITION: 'deposition of',
    Neutral.FLIGHT_LOG: Neutral.FLIGHT_LOG.replace('_', ' '),
    Neutral.PRESSER: Neutral.PRESSER.replace('_', ' '),
    Neutral.RESUMÉ: 'professional resumé',
    Uninteresting.BOOK: 'book titled',
    Uninteresting.TWEET: Uninteresting.TWEET.title(),
}

SHORT_TRUNCATE_CATEGORIES = [
    Uninteresting.ARTICLE,
]


@dataclass(kw_only=True)
class DocCfg(LoggingEntity):
    """
    Encapsulates info about files that needs to be manually configured because it cannot be programmatically inferred.
    Setting the `is_interesting` flag overrides all other considerations when determining a document's
    interestingness (or lack thereof) and also has the additional side effect of causing the entire file to be
    printed (it's the equivalent of setting truncate_to=NO_TRUNCATE).

    Attributes:
        id (str): ID of file
        attached_to_email_id (str, optional): ID of `Email` object this document was an attachment of
        author (Name): Author of the document (if any)
        author_reason (str, optional): Optional explanation of why we are sure this email can be attributed to this author
        author_uncertain (str | bool, optional): Like `author_reason` but for author attributions that aren't 100% confirmed
        background_color (str, optional): set `is_interesting=True` and show in full panel with background color
        category (str, optional): Type of file
        comment (str, optional): Info about this file not worth being in the `note`
        date (str, optional): Parsed to a datetime by timestamp() if it exists
        display_text (str, optional): Replace the contents of this file with this string when it's displayed
        dupe_type (DuplicateType | None): The type of duplicate this file is or its 'duplicate_ids' are
        duplicate_ids (list[str]): IDs of *other* documents that are dupes of this document
        duplicate_of_id (str | None): If this is a dupe the ID of the duplicated file. This file will be suppressed
        highlight_quote (str, optional): text to highlight, also triggers `is_of_interest` and `show_full_panel` to True
        is_interesting (bool | int | None): Force this file to be interesting (or not) if set, int value ranks most intersting
        is_in_chrono (bool | None): Like is_interesting, but only applies with --output-chrono
        is_synthetic (bool): True if this config was generated by the `duplicate_cfgs()` method
        is_valid_for_name_scan (bool): should text of this doc be scanned for names to create biographical panels
        non_participants (list[str]): hacky way to avoid false detection of these names
        note (str, optional): description of what's in this file
        num_preview_chars (int, optional): customize number of preview_chars shown in `OtherFile` tables
        people (list[str]): override `Document.people()` with a fixed set of names (meaning no scan of the text)
        show_full_panel (bool, optional): set `is_interesting=True` and show in a full panel view, not in a table
        show_with_name (str, optional): if set this document will be displayed all with the person specified
        truncate_to (int | tuple[int, int], optional): Number of characters to truncate this email to when displayed
        url (str, optional): URL with more info about this document
        url_link_text (str, optiona): text to show when displaying the `url` link for this document
    """
    id: str
    attached_to_email_id: str | None = None
    author: Name = None
    author_reason: str = ''
    author_uncertain: bool | str = ''
    background_color: str = ''
    category: str = ''
    comment: str = ''
    date: str = ''
    date_uncertain: str | bool = False
    display_text: str = ''
    dupe_type: DuplicateType | None = None
    duplicate_ids: list[str] = field(default_factory=list)
    duplicate_of_id: str | None = None
    entity_names: list[str] = field(default_factory=list)
    highlight_quote: str = ''
    is_interesting: bool | int | None = None  # NOTE: if truthy emails will not be truncated!
    is_in_chrono: bool | None = None
    is_synthetic: bool | None = None
    is_valid_for_name_scan: bool = True
    non_participants: list[str] = field(default_factory=list)  # TODO: this sucks
    note: str = ''
    num_preview_chars: int | None = None
    show_full_panel: bool = False
    show_with_name: str = ''
    truncate_to: int | tuple[int, int] | None = None
    url: str = ''
    url_link_text: str = ''

    def __post_init__(self):
        if self.id in self.duplicate_ids:
            raise ValueError(f"{self.id} is a duplicate of itself!")
        elif 'efta' in self.id:
            self._warn(f"id should not be lowercase: '{self.id}'")
            self.id = self.id.upper()

        self.set_category(self.category)

        # background_color, highlight_quote, or a tuple truncate_to set show_full_panel to true
        if self.background_color or self.highlight_quote or isinstance(self.truncate_to, tuple):
            self.show_full_panel = True

            if self.highlight_quote:
                description_quote = collapse_whitespace(self.highlight_quote.replace('>', ''))
                self.note = join_truthy(self.note, f'{QUOTE_PREFIX}: {quote(description_quote)}', ', ')

        # show_full_panel sets is_interesting=10
        if self.show_full_panel and self.is_interesting is not False:
            self.is_interesting = 10

        if self.duplicate_of_id or self.duplicate_ids:
            self.dupe_type = self.dupe_type or SAME

    @property
    def author_str(self) -> str:
        return self.author or ''

    @property
    def category_txt(self) -> Text:
        """Returns '???' for missing category."""
        from epstein_files.output.highlight_config import styled_category
        return styled_category(self.category)

    @property
    def char_range(self) -> CharRange | None:
        """`truncate_to` as `(0, truncate_to)` tuple if truncate_to is an `int`."""
        if args.whole_file or self.truncate_at in [None, NO_TRUNCATE]:
            return None
        elif isinstance(self.truncate_at, tuple):
            return self.truncate_at
        elif isinstance(self.truncate_at, int):
            return (0, self.truncate_at)
        else:
            raise ValueError(f"{self.id} unknown truncate_at type ({type(self.truncate_at).__name__}, value={self.truncate_at})")

    @property
    def complete_description(self) -> str:
        """String that summarizes what is known about this document."""
        author = f"{self.author} {QUESTION_MARKS}" if self.author and self.author_uncertain else self.author
        preamble = CATEGORY_PREAMBLES.get(self.category, '')
        preamble_separator = ''
        author_separator = ''
        description = ''

        # If description is set at all in one of these if/else checks must be fully constructed
        if self.display_text and not self.has_full_ocr_text_replacement:
            return join_truthy(self.display_text, self.note, ', ')
        if self.category == Uninteresting.BOOK or \
                (self.category == Uninteresting.ACADEMIA and self.author and self.note):
            description = join_truthy(self.note, author, ' by ')  # note reversed args
            description = join_truthy(preamble, description)
        elif self.category == Neutral.FINANCE and self._is_description_a_title:
            author_separator = ' report: '
        elif self.category == Neutral.PRESSER:
            description = join_truthy(preamble, self.note, ' announcing ')  # note reversed args
            description = join_truthy(author, description)
        elif self.category == Interesting.REPUTATION or (self.category == Neutral.LEGAL and 'v.' in self.author_str):
            author_separator = ': '
        elif self.category in [Category.RESUMÉ, Category.TWEET]:
            preamble_separator = 'of' if self.category == Category.RESUMÉ else 'by'
            preamble_separator = preamble_separator.center(3, ' ')

        # Construct standard description from pieces if a custom one has not been created yet
        if not description:
            preamble_author = join_truthy(preamble, author, preamble_separator)
            author_description = join_truthy(author, self.note, author_separator)

            if author and preamble_author.endswith(author) and author_description.startswith(author):
                preamble_author = preamble_author.removesuffix(author).strip()

            description = join_truthy(preamble_author, author_description)

        # TODO: this sucks
        if self.author == INSIGHTS_POD:
            description = join_truthy(description, f"from {ZUBAIR_AND_ANYA}")

        if self.attached_to_email_id:
            description = join_truthy(description, f"attached to email {self.attached_to_email_id}", sep=', ')

        return description

    @property
    def display_preview_txt(self) -> Text | None:
        """Override for the `OtherFile`s table preview column."""
        if self.replacement_preview_text:
            text = join_truthy(self.author, self.replacement_preview_text)
            return Text(f"(Text of {text} {CHECK_LINK_FOR_DETAILS})", 'dim italic')

    @property
    def external_link(self) -> ExternalLink | None:
        """Link to more info about this document (almost unused)."""
        return ExternalLink(self.url, self.url_link_text) if self.url else None

    @property
    def external_link_txt(self) -> Text | None:
        """Link to more info about this document (almost unused)."""
        return self.external_link.domain_link if self.external_link else None

    @property
    def has_any_info(self) -> bool:
        """True if either `author` or `description` is truthy."""
        return bool(self.note or self.author)

    @property
    def has_full_ocr_text_replacement(self) -> bool:
        """`display_text` longer than other_files_preview_chars is considered a drop in replacement, not a short description."""
        return bool(self.display_text and len(self.display_text) > MobileConfig.other_files_preview_chars)

    @property
    def note_txt(self) -> Text | None:
        """Add formatting to `self.complete_description`."""
        if self.complete_description:
            txt = Text(self.complete_description, NOTE_STYLE)

            if self.external_link_txt:
                txt.append(' ').append(self.external_link_txt)

            from epstein_files.output.epstein_highlighter import non_epstein_highlighter
            return non_epstein_highlighter(txt)

    @property
    def replacement_preview_text(self) -> str:
        """Returns a string if `self.display_text` exists and is not too long."""
        return self.display_text if self.display_text and not self.has_full_ocr_text_replacement else ''

    @property
    def text_highlighter(self) -> 'EpsteinHighlighter':
        """Use a custom highlighter that also colors `self.highlight_quote` string if set."""
        from epstein_files.output.epstein_highlighter import highlighter, temp_highlighter

        if self.highlighted_pattern:
            return temp_highlighter(self.highlighted_pattern)
        else:
            return highlighter

    @property
    def highlighted_pattern(self) -> str | None:
        """Regex pattern that matches `self.highlight_quote` string, allowing for line breaks etc."""
        return re.escape(self.highlight_quote).replace(r'\ ', r"\s+") if self.highlight_quote else None

    @property
    def is_doj_file(self) -> bool:
        return is_doj_file(self.id)

    @property
    def is_empty(self) -> bool:
        has_truthy_field = any([v for k, v in asdict(self).items() if k not in ['id', 'is_valid_for_name_scan']])
        return self.is_valid_for_name_scan is True and not has_truthy_field

    @property
    def is_excerpt(self) -> bool:
        return isinstance(self.truncate_to, tuple)

    @property
    def is_house_file(self) -> bool:
        return not self.is_doj_file

    @property
    def is_of_interest(self) -> bool | None:
        """
        `self.is_interesting` value takes precedence. If it's not set apply the rules below.
        Defaults to True for HOUSE_OVERSIGHT files w/out info, None for DOJ files.
        Returns None (not False) if there's no firm decision, leaving `Document` classes
        to do any other checks they might want to.

                [+] = interesting  /  - = uninteresting

            [+] is_interesting (+ is_in_chrono if --output-chrono in effect)
            [+] highlight_quote is set
            [+] author in PERSONS_OF_INTEREST
            [+] category is in Interesting enum
            [+] having no author/description *if* HOUSE_OVERSIGHT
             -  duplicates
             -  known email attachments
             -  is_interesting == False (not None)
             -  finance category with any author (more broadly "Neutral categories can check bespoke additional conditions")
             -  category in Uninteresting enum
             -  description text starts with one of UNINTERESTING_PREFIXES
        """
        if self.duplicate_of_id:
            return False
        elif self.attached_to_email_id and (args.output_chrono or args.output_emails):  # TODO: not sure whether it's good always exclude attachments
            return False
        elif args.output_chrono and self.is_in_chrono is not None:
            return self.is_in_chrono
        elif self.is_interesting is not None:
            return bool(self.is_interesting)
        elif self.highlight_quote:
            return True
        # author related checks  # NOTE: this only applies to configured authors or derived_cfg! so other files but not most emails
        elif self.author and self.author in PERSONS_OF_INTEREST:
            return True
        # category field checks
        elif is_interesting(self.category):
            return True
        elif self.category == Category.FINANCE and self.author is not None:  # TODO: boothbay and other pitch decks should be of interest
            return False
        elif is_uninteresting(self.category):
            return False
        # description field checks
        elif any (self.note.startswith(pfx) for pfx in UNINTERESTING_PREFIXES):
            return False

        return None

    @property
    def is_very_interesting(self) -> bool:
        return isinstance(self.is_interesting, int)

    @property
    def metadata(self) -> Metadata:
        metadata = {k: v for k, v in asdict(self).items() if v and k not in NON_METADATA_FIELDS}

        if self.is_interesting is False:
            metadata['is_interesting'] = False  # Configured False is valid signal

        return metadata

    @property
    def names(self) -> list[str]:
        """Names configured for this document. Overloaded in subclass to add recipients."""
        return [self.author] if self.author else []

    @property
    def timestamp(self) -> datetime | None:
        if self.date and (parsed_dt := coerce_utc_strict(parse(self.date))):
            self._debug_log(f"parsed {parsed_dt.isoformat()} from date='{self.date}'")
            return parsed_dt

    @property
    def truncate_at(self) -> int | CharRange | None:
        """The number of chars to show when printing this document."""
        if self.truncate_to:
            return self.truncate_to
        elif self.is_interesting:
            return NO_TRUNCATE
        elif self.category in SHORT_TRUNCATE_CATEGORIES:
            return SHORT_TRUNCATE_TO

    @property
    def truthy_props(self) -> dict[str, bool | str | None]:
        props = {k: v for k, v in asdict(self).items() if v or (is_bool_prop(k) and v is False)}

        if self.is_of_interest is not None:
            if self.is_of_interest == props.get('is_interesting'):
                props['is_of_interest'] = props.pop('is_interesting')  # Remove is_interesting, just keep is_of_interest
            else:
                props['is_of_interest'] = self.is_of_interest

        if self.complete_description:
            description_pieces = without_falsey([self.author, self.note])

            # Avoid showing complete_description if it's just the author or description and other prop doesn't exist
            if len(description_pieces) != 1 or description_pieces[0] != self.complete_description:
                props['complete_description'] = self.complete_description

        if (category_txt := self.category_txt):
            if category_txt.plain == props.get('category'):
                props.pop('category')  # Leave only the colored version of category_txt

            # Only add ??? for non-email, non immesage. # TODO: this sucks
            if category_txt.plain == QUESTION_MARKS:
                if 'recipients' not in dir(self):
                    props['category_txt'] = category_txt
            else:
                props['category_txt'] = category_txt

        if self.timestamp:
            props['timestamp'] = self.timestamp

            if 'date' in props:
                props.pop('date')

        if props.get('dupe_type') == SAME:
            props.pop('dupe_type')

        return props

    @property
    def _class_name(self) -> str:
        return type(self).__name__

    @property
    def _identifier(self) -> str:
        """Required `LoggingEntity` abstract method."""
        return self.id

    @property
    def _is_description_a_title(self) -> bool:
        """True if first char is uppercase or a quote."""
        if not (self.author and self.note):
            return False
        elif self.category not in [Category.ACADEMIA, Category.BOOK, Category.FINANCE]:
            return False
        elif self.category == Category.FINANCE and self.author not in FINANCIAL_REPORTS_AUTHORS:
            return False

        first_char = self.note[0]
        return first_char.isupper() or first_char in ["'", '"']

    def duplicate_cfgs(self) -> Generator[Self, None, None]:
        """Create synthetic `DocCfg` objects that set the 'duplicate_of_id' field to point back to this object."""
        for id in self.duplicate_ids:
            dupe_cfg = deepcopy(self)
            dupe_cfg.id = id
            dupe_cfg.duplicate_ids = []
            dupe_cfg.duplicate_of_id = self.id
            dupe_cfg.dupe_type = self.dupe_type
            dupe_cfg.is_synthetic = True
            yield dupe_cfg

    def set_category(self, category: str) -> None:
        """Update the title if we changed to a category that allows titling (books, academia, finance)."""
        self.category = category.lower().strip()

        if not self.category:
            return
        elif not is_category(self.category):
            self._warn(f"'{self.category}' does not appear to be a valid category")

        if self.category == Category.FLIGHT_LOG and not self.display_text:
            self.display_text ='flight log'

        if self.category == Uninteresting.BOOK:
            self.is_valid_for_name_scan = False

        self.note = quote(self.note) if self._is_description_a_title else self.note

    def _props_strs(self) -> list[str]:
        props = []
        add_prop = lambda f, value: props.append(f"{f.name}={value}")

        for _field in sorted(fields(self), key=lambda f: FIELD_SORT_KEY.get(f.name, f.name)):
            value = getattr(self, _field.name)

            if _field.name in ['actual_text'] or is_bool_prop(_field.name):  # fields can be False or None or ''
                if value is not None:
                    add_prop(_field, str(value))
            elif not value or (_field.name == 'dupe_type' and value == 'same'):
                continue
            elif _field.name == AUTHOR:
                add_prop(_field, constantize_name(str(value)) if args.constantize else f"'{value}'")
            elif _field.name == 'recipients':
                recipients_str = str([constantize_name(r) if (args.constantize and r) else r for r in value])
                add_prop(_field, recipients_str.replace("'", '') if args.constantize else recipients_str)
            elif _field.name == 'truncate_to' and value == NO_TRUNCATE:
                add_prop(_field, 'NO_TRUNCATE')
            elif isinstance(value, str):
                if "'" in value:
                    value = '"' + value.replace('"', r'\"') + '"'
                else:
                    value = "'" + value.replace("'", r'\'') + "'"

                add_prop(_field, value)
            else:
                add_prop(_field, str(value))

        return props

    def __rich__(self) -> Text:
        return Text(', ').join([Text(p) for p in self._props_strs()])

    def __repr__(self) -> str:
        props = self._props_strs()
        type_str = f"{self._class_name}("
        single_line_repr = type_str + ', '.join(props) + f')'

        if len(single_line_repr) < MAX_REPR_LINE_LENGTH or \
                (self.comment and 'is_fwded_article' in dir(self) and getattr(self, 'is_fwded_article')):
            repr_str = single_line_repr
        else:
            repr_str = f"{type_str}{INDENT_NEWLINE}" + INDENTED_JOIN.join(props)
            repr_str += ',' if props else ''
            repr_str += '\n)'

        if args.constantize:
            repr_str = INDENT + INDENT_NEWLINE.join(repr_str.split('\n'))
            return repr_str.replace(',,', ',').replace(',),', '),').replace(',),', '),')
        else:
            return repr_str

    @classmethod
    def create_or_set_prop(cls, ids: list[str], existing_cfgs: Sequence['DocCfg'], prop: str, val: Any) -> None:
        """If a record exists in `existing_cfgs` update it, otherwise create new and append."""
        cfg_dict = {cfg.id: cfg for cfg in existing_cfgs}
        created = updated = 0

        for id in ids:
            if (cfg := cfg_dict.get(id)):
                assert getattr(cfg, prop) is None, f"Can't overwrite existing '{prop}' value for {cfg}"
                setattr(cfg, prop, val)
                updated += 1
            else:
                existing_cfgs.append(cls(id=id, **{prop: val}))
                created += 1

        logger.info(f"Created {created} {cls.__name__} with {prop}={val}, updated {updated} existing.")

    @classmethod
    def set_categories(cls, cfgs: Sequence['DocCfg'], category: str | Path) -> None:
        """Set the `category` property for all `cfgs`."""
        category = category.stem if isinstance(category, Path) else category
        logger.debug(f"Setting category for {len(cfgs)} configs to '{category}'")

        for cfg in cfgs:
            cfg.set_category(cfg.category or category)



"""
String manipulation.
"""
import re
from datetime import datetime
from typing import Any

from inflection import underscore

from epstein_files.util.constant.strings import AMPERSAND_CHAR_GROUP, QUESTION_MARKS_REGEX

EMOJI_REGEX = re.compile(r"(?:^|\s)([:;=][-^]?[oODP()]|[oO()][-^]?[:=])(?=$|\s)")
INTEGER_REGEX = re.compile(r'^\d+$')
MULTINEWLINE_REGEX = re.compile(r"\n{2,}")
MULTISPACE_REGEX = re.compile(" +")
PDFALYZER_IMAGE_PANEL_REGEX = re.compile(r"\n╭─* Page \d+, Image \d+.*?╯\n?", re.DOTALL)
TIMESTAMP_SECONDS_REGEX = re.compile(r":\d{2}(\.\d+)?([-+]\d{2}:\d{2})?$")
WHITESPACE_REGEX = re.compile(r"\s{2,}|\t|\n", re.MULTILINE)

WHITESPACE_CHAR = r"[-_.\s]*"
DATE_LENGTH = len('2025-05-05')
DOUBLESPACE_IF_LINE_LEN_OVER = 130
DOUBLESPACE_IF_LONG_LINE_PCT = 0.5

capitalize_first = lambda s: s[0].upper() + s[1:]
capture_group_marker = lambda label: fr"?P<{label}>"
collapse_newlines = lambda text: MULTINEWLINE_REGEX.sub('\n\n', text)
collapse_spaces = lambda s: MULTISPACE_REGEX.sub(' ', s)
collapse_whitespace = lambda s: WHITESPACE_REGEX.sub(' ', s).strip()
constantize = lambda s: underscore(s.upper())
is_bool_prop = lambda prop: prop.startswith('is_')
is_integer = lambda s: bool(INTEGER_REGEX.match(s))
join_patterns = lambda patterns: '|'.join(patterns)
iso_timestamp = lambda dt: dt.isoformat().replace('T', ' ')
strip_pdfalyzer_panels = lambda s: PDFALYZER_IMAGE_PANEL_REGEX.sub('', s)
timestamp_str = lambda dt: dt.isoformat()[0:19]
timestamp_human = lambda dt: timestamp_str(dt).replace('T', ' ')

# regexes
or_equal_sign_char_group = lambda s: f"[{s}=]"  # DataSet 11 has a lot of random '=' replacing characters


def as_pattern(s: str) -> str:
    """Replace spaces with regex pattern for whitespace and ampersands with common OCR fails."""
    s = collapse_spaces(s).replace('@', AMPERSAND_CHAR_GROUP)
    return s if '?<!' in s else s.replace(' ', WHITESPACE_CHAR)


def doublespace_lines(s: str) -> str:
    """Doublespace \n chars if s has a high pct of long lines."""
    lines = s.split('\n')
    long_lines = [line for line in lines if len(line) > DOUBLESPACE_IF_LINE_LEN_OVER]

    if (len(long_lines) / len(lines)) > DOUBLESPACE_IF_LONG_LINE_PCT:
        return s.replace('\n', '\n\n')
    else:
        return s


def extract_emojis(s: str) -> list[str]:
    return [m.group(1) for m in EMOJI_REGEX.finditer(s)]


def has_line_starting_with(s: str | list[str], pfxes: str | list[str], limit: int | None = None) -> bool:
    lines = s.split('\n') if isinstance(s, str) else s
    lines = lines[0:(limit or len(lines))]

    for line in lines:
        if any(line.startswith(pfx) for pfx in list(pfxes)):
            return True

    return False


def indented(text: str | list[str], spaces: int = 4, prefix: str = '') -> str:
    line_prefix = (' ' * spaces) + prefix
    lines = text.split('\n') if isinstance(text, str) else text
    return line_prefix + f"\n{line_prefix}".join(lines)


def join_truthy(prefix: str | None, suffix: str | None, sep: str = '') -> str:
    """Join two strings but only if they are not empty."""
    sep = sep or (' ' if prefix and suffix else '')
    parts = [p.strip() for p in [prefix, suffix] if p and p.strip()]
    return sep.join(parts)


def prop_str(prop: Any) -> Any:
    return quote(str(prop)) if isinstance(prop, (datetime, str)) else str(prop)


def quote(s: str, try_single_quote_first: bool = False) -> str:
    if '"' in s and "'" in s:
        return s
    elif s and s[0] in ['"', "'"]:
        return s
    elif try_single_quote_first:
        return f"'{s}'" if "'" not in s else f'"{s}"'
    else:
        return f"'{s}'" if '"' in s else f'"{s}"'


def remove_question_marks(name: str):
    return QUESTION_MARKS_REGEX.sub('', name).strip()


def snip_msg(msg: str) -> str:
    return f'<...{msg}...>'


def starred_header(msg: str, num_stars: int = 7, num_spaces: int = 2) -> str:
    """String like '  *** Title Msg ***  '."""
    stars = '*' * num_stars
    spaces = ' ' * num_spaces
    return f"{spaces}{stars} {msg} {stars}{spaces}"


def timestamp_without_seconds(dt: datetime) -> str:
    return TIMESTAMP_SECONDS_REGEX.sub('', str(dt))


def timestamp_without_zero_hour(dt: datetime) -> str:
    """Remove the time part of a datetime string if it's 00:00:00, otherwise keep it."""
    return dt.strftime(r'%Y-%m-%d %H:%M:%S').removesuffix(' 00:00:00')



import json
import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from typing import ClassVar, cast, TypeVar

from dateutil.parser import parse
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, EXCERPT_STYLE
from epstein_files.documents.documents.categories import Uninteresting
from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, NO_TRUNCATE, SHORT_TRUNCATE_TO, DebugDict, Metadata
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.emails.constants import *
from epstein_files.documents.emails.email_parts import EmailParts
from epstein_files.documents.emails.email_header import (EMAIL_SIMPLE_HEADER_REGEX,
     EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX, EmailHeader)
from epstein_files.documents.emails.emailers import IDENTIFYING_REGEXES, IDENTIFIER_FALSE_ALARMS, extract_emailer_names
from epstein_files.documents.other_file import OtherFile
from epstein_files.people.entity import Entity, EntityScanArg
from epstein_files.people.interesting_people import EMAILERS_OF_INTEREST_SET
from epstein_files.people.names import sort_names
from epstein_files.output.epstein_highlighter import highlighter
from epstein_files.output.highlight_config import HIGHLIGHTED_NAMES, get_style_for_name
from epstein_files.output.html.builder import table_to_html
from epstein_files.output.html.positioned_rich import to_em
from epstein_files.output.layout_elements.file_display import FileDisplay, JustifyMethod
from epstein_files.output.rich import DEFAULT_TABLE_KWARGS, build_table, styled_key_value
from epstein_files.util.constant.strings import APPEARS_IN, ARCHIVE_LINK_COLOR, REDACTED, TIMESTAMP_DIM, OcrRepair
from epstein_files.util.constant.urls import URL_SIGNIFIERS
from epstein_files.util.constants import CONFIGS_BY_ID
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import (AMERICAN_TIME_REGEX, TIMEZONE_INFO, CharRange, coerce_utc, flatten,
     prefix_keys, uniq_sorted, uniquify, without_falsey)
from epstein_files.util.helpers.rich_helpers import enclose
from epstein_files.util.external_link import join_texts, link_text_obj
from epstein_files.util.helpers.string_helper import capitalize_first, collapse_newlines, is_bool_prop, quote, strip_pdfalyzer_panels
from epstein_files.util.logging import logger

# Email bod regexes
BAD_FIRST_LINE_REGEX = re.compile(r'^(>>|Grant_Smith066474"eMailContent.htm|LOVE & KISSES)$')
BAD_LINE_REGEX = re.compile(r'^([>=];?|[>»]*=20|\d{1,2}|PAGE INTENTIONALLY LEFT BLANK|Classification: External Communication|Hide caption|Importance:?\s*High|[iI,•]|[1i] (_ )?[il]|, [-,]|L\._|_filtered|si.nature.asc|.*(yiv0232|font-family:|margin-bottom:).*)$')
BAD_SUBJECT_CONTINUATIONS = ['orwarded', 'Hi ', 'Sent ', 'AmLaw', 'Original Message', 'Privileged', 'Sorry', '---']
LINK_LINE_REGEX = re.compile(r"^[>• ]*htt")
LINK_LINE2_REGEX = re.compile(r"^[-\w.%&=/]{5,}$")
QUOTED_REPLY_LINE_REGEX = re.compile(r'(\nFrom:(.*)|wrote:)\n', re.IGNORECASE)
REPLY_TEXT_REGEX = re.compile(rf"^(.*?){REPLY_LINE_PATTERN}", re.DOTALL | re.IGNORECASE | re.MULTILINE)
XML_PLIST_REGEX = re.compile(r"[<=]?\?xml version.*</(plist|xml)>", re.DOTALL)

# Timestamp regexes
BAD_TIMEZONE_REGEX = re.compile(fr'\((UTC|GMT\+\d\d:\d\d)\)|{REDACTED}')
DATE_HEADER_REGEX = re.compile(r'(?:Date|Sent):? +(?!by|from|to|via)([^\n]{6,})\n')
TIMESTAMP_LINE_REGEX = re.compile(r"\d+:\d+")

# numbers
MAX_NUM_HEADER_LINES = 14
MAX_QUOTED_REPLIES = 1
NUM_LINES_TO_REPAIR_HEADERS = 6
NUM_WORDS_IN_LAST_QUOTE = 6

# TODO: Copy display_text?
DERIVED_CFG_PROPS_TO_COPY = [
    'author_uncertain',
    'category',
    'is_in_chrono',
    'is_interesting',
]

# Junk mail
JUNK_EMAILERS = [
    entity.name
    for junk_hg in HIGHLIGHTED_NAMES if junk_hg.label == Uninteresting.JUNK
    for entity in junk_hg.entities
]

BCC_LISTS = JUNK_EMAILERS + MAILING_LISTS
TRUNCATE_EMAILS_BY = BCC_LISTS + TRUNCATE_EMAILS_FROM
EMAILERS_TO_ALWAYS_TRUNCATE = set(BCC_LISTS + TRUNCATE_EMAILS_BY)
REWRITTEN_HEADER_MSG = "(janky OCR header fields were prettified, check source if something seems off)"

# TODO: add other forward patterns
REPLY_SPLITTERS = [f"{field}:" for field in COMMON_HEADER_FIELDS] + [
    '********************************',
    'Begin forwarded message',
]

OCR_REPAIRS: OcrRepair = {
    # '­': '-',
    # '‐': '-',
    '-­‐': '-',  # TODO: weird hyphens in 027004 and other files that rich doesn't handle well
    re.compile('»'): '>>',
    re.compile(r'grnail\.com'): 'gmail.com',
    'Newsmax. corn': 'Newsmax.com',
    # These 3 must come in this order!
    re.compile(r'([/vkT]|Ai|li|(I|7)v)rote:'): 'wrote:',
    re.compile(r"([<.=_HIM][<>.=_HIM14]{5,}[<>.=_HIM]|MOMMINNEMUMMIN) *(wrote:?)?"): rf"{REDACTED} \2",
    re.compile(r"([,<>_]|AM|PM)\n(>)? ?wrote:?"): r'\1\2 wrote:',
    # Headers
    re.compile(r"^From "): 'From: ',  # first line only
    re.compile(r"^((From|To):? ?)[_1.]{5,}", re.MULTILINE): rf"\1: {REDACTED}",  # Redacted email addresses
    re.compile(fr"^(From|To) ?{REDACTED}", re.MULTILINE): fr"\1: {REDACTED}",
    re.compile(r"I ?(od|nl)ine-Images:"): 'Inline-Images:',
    re.compile(r"^((?:B?cc|To):.*)\n(>?;.*)", re.IGNORECASE | re.MULTILINE): r'\1 \2',
    re.compile(r"^(Sent|Subject) (?![Ff]rom|on|using|[Rr]emote|[Vv]ia|with)", re.MULTILINE): r'\1: ',
    re.compile(r"^Subject[.•]{,2} ", re.MULTILINE): 'Subject: ',
    re.compile(r"^(Forwarded|Original) Message$", re.IGNORECASE | re.MULTILINE): r"--- \1 Message ---",  # Make forward lines match our highlight
    # Excessive quote chars
    re.compile(r"wrote:\n[>»]+(\n[>»]+)"): r"wrote:\1",
    re.compile(r"(^[>»]+\n){2,}", re.MULTILINE): r"\1",
    re.compile(r"\n[<>I ]*wrote:"): ' wrote:',
    # HTML garbage
    '&lt;': '<',
    '<mime-attachment.gif>': '',
    re.compile(r'(fi|[&S5d])gt;'): '>',
    re.compile(r'[</=]{,3}(cl|d)?iv>|&[=n]bs[=p];|<=span>|=C\d+\b'): '',
    re.compile(r'^O= ', re.MULTILINE): 'On ',
    re.compile(r"<?=/?[bru]>"): '',
    re.compile(r'(^|\s)[<=][AC]\d+[=>]?', re.MULTILINE): r'\1',
    re.compile(r'[<=]=?/?(br|d?iv)( (class|id)="\w+")?>'): '',
    re.compile(r"\n<mailt.:?(.{,25})[»>]\s*wrote"): r'\1 wrote',
    re.compile(r"^--\w+-- (conversation-id|date-last-viewed).*(flags|remote-id\s\d+)(\s*\d{6,}.*remote-id.*\d+)?", re.MULTILINE): '',
    re.compile(r"<mailto:([-\w=.@]+)[»>]"): r'\1',
    # Names / email addresses
    r'Alireza lttihadieh': ALIREZA_ITTIHADIEH,
    r'bamaby': 'barnaby',
    r'Miroslav Laj6ak': MIROSLAV_LAJCAK,
    r'Ross G°w': ROSS_GOW,
    r'Torn Pritzker': TOM_PRITZKER,
    re.compile(r"( [AP]M,)\s+wrote:$", re.MULTILINE): r'\1 <REDACTED> wrote:',
    re.compile(r' Banno(r]?|\b)'): ' Bannon',
    re.compile(r"\bBamaby\b"): 'Barnaby',
    re.compile(r'gmax ?[1l] ?[@g]ellmax.c ?om'): 'gmax1@ellmax.com',
    re.compile(r"[ijlp']ee[vy]acation[©@a(&,P ]{1,3}g?mail.com"): 'jeevacation@gmail.com',
    'gyahoo.com': '@yahoo.com',
    # Signatures
    'Envoy& de': 'Envoyé de',
    'Envoye avec BlackBerry° d': 'Envoye avec BlackBerry® d',
    'from Samsung Mob.le': 'from Samsung Mobile',
    'gJeremyRubin': '@JeremyRubin',
    'Mail for i Phone': 'Mail for iPhone',
    'Sent from Mabfl': 'Sent from Mobile',  # NADIA_MARCINKO signature bad OCR
    'twitter glhsummers': 'twitter @lhsummers',
    re.compile(r'Blac[il]cBerry'): 'BlackBerry',
    'BlackBerry by AT &T': 'BlackBerry by AT&T',  # Must come after previous
    'BlackBerry from T- Mobile': 'BlackBerry from T-Mobile',
    re.compile(r"[cC]o-authored with i ?Phone auto-correct"): "Co-authored with iPhone auto-correct",
    re.compile(r"from my ['!()=]([Pp]hone)"): r'from my i\1',
    re.compile(r'from my BlackBerry[0°] wireless device'): 'from my BlackBerry® wireless device',
    re.compile(r'^INW$', re.MULTILINE): REDACTED,
    re.compile(r'Sent from one of my many test mobile devices while on the go and changing the world\s+:\s+so\s+my\s+apologies for any typos'): 'Sent from one of my many test mobile devices while on the go and changing the world : so my apologies for any typos',
    re.compile(r"twitter\.com[i/][lI]krauss[1lt]"): "twitter.com/lkrauss1",
    # links
    'classified-intelligence-\nmichael-flynn-trump': 'classified-intelligence-michael-flynn-trump',
    'on-accusers-rose-\nmcgowan/ ': 'on-accusers-rose-\nmcgowan/\n',
    'woody-allen-jeffrey-epsteins-\nsociety-friends-close-ranks/ ---': 'woody-allen-jeffrey-epsteins-society-friends-close_ranks/\n',
    ' https://www.theguardian.com/world/2017/may/29/close-friend-trump-thomas-barrack-\nalleged-tax-evasion-italy-sardinia?CMP=share btn fb': '\nhttps://www.theguardian.com/world/2017/may/29/close-friend-trump-thomas-barrack-alleged-tax-evasion-italy-sardinia?CMP=share_btn_fb',
    'search-for-secret-putin-\nfortune.html': 'search-for-secret-putin-fortune.html',
    re.compile(r"[h=][t=][t=][p=]://\s*"): 'http://',
    re.compile(r"([h=][t=][t=][op=][s=]|Imps )://\s*"): 'https://',
    re.compile(r'timestopics/people/t/landon jr thomas/inde\n?x\n?\.\n?h\n?tml'): 'timestopics/people/t/landon_jr_thomas/index.html',
    re.compile(r" http ?://www. ?dailymail. ?co ?.uk/news/article-\d+/Troub ?led-woman-history-drug-\n?us ?e-\n?.*html"): '\nhttp://www.dailymail.co.uk/news/article-3914012/Troubled-woman-history-drug-use-claimed-assaulted-Donald-Trump-Jeffrey-Epstein-sex-party-age-13-FABRICATED-story.html',
    re.compile(r"http.*steve-bannon-trump-tower-\n?interview-\n?trumps-\n?strategist-plots-\n?new-political-movement-948747"): "\nhttp://www.hollywoodreporter.com/news/steve-bannon-trump-tower-interview-trumps-strategist-plots-new-political-movement-948747",
    # Subject lines
    r'Priebus, used\nprivate email accounts for': 'Priebus, used private email accounts for',
    r"avoids testimony from alleged\nvictims": "avoids testimony from alleged victims",
    r"Subject; RE": "Subject: RE",
    re.compile(r"as Putin Mayhem\sTests\sPresident's\sGrip\son\sGOP"): "as Putin Mayhem Tests President's Grip on GOP",
    re.compile(r"deadline re Mr Bradley Edwards vs Mr\s*Jeffrey Epstein", re.I): "deadline re Mr Bradley Edwards vs Mr Jeffrey Epstein",
    re.compile(r"Following Plea That Implicated Trump -\s*https://www.npr.org/676040070", re.I): "Following Plea That Implicated Trump - https://www.npr.org/676040070",
    re.compile(r"for Attorney General -\s+Wikisource, the"): r"for Attorney General - Wikisource, the",
    re.compile(r"JUDGE SWEET\s+ALLOWING\s+STEVEN\s+HOFFENBERG\s+TO\s+TALK\s+WITH\s+THE\s+TOWERS\s+VICTIMS\s+TO\s+EXPLAIN\s+THE\s+VICTIMS\s+SUI\n?T\s+FILING\s+AGAINST\s+JEFF\s+EPSTEIN"): "JUDGE SWEET ALLOWING STEVEN HOFFENBERG TO TALK WITH THE TOWERS VICTIMS TO EXPLAIN THE VICTIMS SUIT FILING AGAINST JEFF EPSTEIN",
    re.compile(r"Lawyer for Susan Rice: Obama administration '?justifiably concerned' about sharing Intel with\s*Trump team -\s*POLITICO", re.I): "Lawyer for Susan Rice: Obama administration 'justifiably concerned' about sharing Intel with Trump team - POLITICO",
    re.compile(r"PATTERSON NEW\s+BOOK\s+TELLING\s+FEDS\s+COVER\s+UP\s+OF\s+BILLIONAIRE\s+JEFF\s+EPSTEIN\s+CHILD\s+RAPES\s+RELEASE\s+DATE\s+OCT\s+10\s+2016\s+STEVEN\s+HOFFENBERG\s+IS\s+ON\s+THE\s+BOOK\s+WRITING\s+TEAM\s*!!!!"): "PATTERSON NEW BOOK TELLING FEDS COVER UP OF BILLIONAIRE JEFF EPSTEIN CHILD RAPES RELEASE DATE OCT 10 2016 STEVEN HOFFENBERG IS ON THE BOOK WRITING TEAM !!!!",
    re.compile(r"PROCEEDINGS FOR THE ST THOMAS ATTACHMENT OF\s*ALL JEFF EPSTEIN ASSETS"): "PROCEEDINGS FOR THE ST THOMAS ATTACHMENT OF ALL JEFF EPSTEIN ASSETS",
    re.compile(r"Subject:\s*Fwd: Trending Now: Friends for three decades"): "Subject: Fwd: Trending Now: Friends for three decades",
    # Russian
    # Bocmpeceime
    re.compile(r'^B(TOpHHK|oc[xKm]pec\w+)', re.MULTILINE | re.IGNORECASE): 'Вторник',
    re.compile(r"^Cpe\w{2,},", re.MULTILINE | re.IGNORECASE): 'среда,',
    re.compile(r"^Cy66.rr?a", re.MULTILINE | re.IGNORECASE): 'суббота',
    re.compile(r"^flo\w{7,8}[HKhkw],", re.MULTILINE): 'понедельник,',
    re.compile(r"^TeMa:", re.MULTILINE | re.IGNORECASE): 'Тема:',
    re.compile(r"^KoMy:", re.MULTILINE | re.IGNORECASE): 'Кому:',
    re.compile(r"^Aara:", re.MULTILINE | re.IGNORECASE): 'Дата:',
    re.compile(r"^[O0]T:", re.MULTILINE | re.IGNORECASE): 'От:',
    # XML footers
    re.compile('<[iI] ?nteger>'): '<integer>',
    # Misc
    'AVG°': 'AVGO',
    'Saw Matt C with DTF at golf': 'Saw Matt C with DJT at golf',
    re.compile(r"[i. ]*Privileged[- ]*Redacted[i. ]*"): '<PRIVILEGED - REDACTED>',
    re.compile(r"SONY ?(Court|Judge|(, |/)NY)", re.IGNORECASE): r'SDNY \1',
}

METADATA_FIELDS = [
    'attachments',
    'attachment_file_ids',
    'is_junk_mail',
    'is_mailing_list',
    'extracted_recipients',
    'recipients',
    'sent_from_device',
]

DEBUG_PROPS = METADATA_FIELDS + [
    'attachment_file_ids',
    'is_persons_first_email',
    'is_word_count_worthy',
]


@dataclass
class Email(Communication):
    """
    Attributes:

        attached_docs (list[OtherFile]): any attachments that exist as `OtherFile` objects
        actual_text (str): Best effort at the text *actually* sent in this email excluding quoted replies and forwards
        derived_cfg (EmailCfg): EmailCfg that was built instead of coming from `CONFIGS_BY_ID`
        is_persons_first_email (bool): is this the first email we have for any of the participants (sender + recipients)
        sent_from_device (str, optional): "Sent from my iPhone" style signature (if it exists).
        signature_substitution_counts (dict[str, int]): Number of times a signature was replaced with
            <...snipped...> per name

        _header (EmailHeader): Header data extracted from the text (from/to/sent/subject etc).
        _line_merge_arguments (list[tuple[int] | tuple[int, int]]): preconfigured list of line merges that will fix up
            files from the HOUSE_OVERSIGHT_ collection in memory while leaving the source files untouched
        _was_split_up (bool, optional): True if this file Email was one of the big ones that was split into pieces
            and thus should generally be hidden / not shown
    """
    attached_docs: list[OtherFile] = field(default_factory=list)
    actual_text: str = field(init=False)
    derived_cfg: EmailCfg | None = None
    is_persons_first_email: bool = False
    sent_from_device: str | None = None
    signature_substitution_counts: dict[str, int] = field(default_factory=dict)  # defaultdict breaks asdict :(
    _header: EmailHeader | None = None
    _line_merge_arguments: list[tuple[int] | tuple[int, int]] = field(default_factory=list)
    _was_split_up: bool = False

    # Class variable logging how many headers we prettified while printing, kind of janky
    rewritten_header_ids: ClassVar[set[str]] = set([])

    def __post_init__(self):
        super().__post_init__()
        self.actual_text = self._extract_actual_text()
        self.sent_from_device = self._sent_from_device()

        # Scan for any identifiers we may have missed that could unredact this email
        for regex, contact in IDENTIFYING_REGEXES.items():
            if regex.search(self.text) and contact.name not in self.participants and self.file_id not in IDENTIFIER_FALSE_ALARMS:
                self._warn(f"Found known identifier for {contact.name} in email where they are not an identified participant")

    @property
    def attachment_file_ids(self) -> list[str]:
        """Strings in the Attachments: field in the header, split by semicolon."""
        return [a.file_id for a in self.attached_docs]

    @property
    def attachments(self) -> list[str]:
        """Strings in the Attachments: and Inline-Images: fields in the header, split by semicolon."""
        return self.header.all_attachments

    @property
    def char_range_to_display(self) -> CharRange | None:
        """Override superclass to decide how many chars we should limit the dislpay of this email to."""
        quote_cutoff = self._idx_of_nth_quoted_reply()  # Trim if there's many quoted replies
        includes_truncate_term = next((term for term in TRUNCATE_TERMS if term in self.text), None)
        num_chars: int | None = None

        if args.whole_file or self._config.truncate_at == NO_TRUNCATE:
            return None
        elif args.truncate:
            num_chars = args.truncate
        elif self._config.char_range:
            return self._config.char_range
        elif self.author in TRUNCATE_EMAILS_BY \
                or any([self.is_from_or_to(n) for n in TRUNCATE_EMAILS_FROM_OR_TO]) \
                or includes_truncate_term:
            num_chars = min(quote_cutoff or DEFAULT_TRUNCATE_TO, SHORT_TRUNCATE_TO)
        else:
            if quote_cutoff and quote_cutoff < DEFAULT_TRUNCATE_TO:
                trimmed_words = self.text[quote_cutoff:].split()

                # TODO this attempt to include <snipped> msgs in the truncated text kind of sucks
                if '<...snipped' in trimmed_words[:NUM_WORDS_IN_LAST_QUOTE]:
                    num_trailing_words = 0
                elif trimmed_words and trimmed_words[0] in ['From:', 'Sent:']:
                    num_trailing_words = NUM_WORDS_IN_LAST_QUOTE
                else:
                    num_trailing_words = NUM_WORDS_IN_LAST_QUOTE

                if trimmed_words:
                    last_quoted_text = ' '.join(trimmed_words[:num_trailing_words])
                    num_chars = quote_cutoff + len(last_quoted_text) + 1 # Give a hint of the next line
                else:
                    num_chars = quote_cutoff
            elif self.length > DEFAULT_TRUNCATE_TO:
                num_chars = DEFAULT_TRUNCATE_TO

            # Always print whole email for 1st email for actual people
            if num_chars and self.is_persons_first_email and self.is_word_count_worthy:
                self._log(f"Overriding cutoff {num_chars} for first email")
                num_chars = None

        log_args = {
            'num_chars': num_chars,
            'author_truncate': self.author in TRUNCATE_EMAILS_BY,
            'is_fwded_article': self.is_fwded_article,
            'is_quote_cutoff': quote_cutoff == num_chars,
            'includes_truncate_term': json.dumps(includes_truncate_term) if includes_truncate_term else None,
            'quote_cutoff': quote_cutoff,
            'is_persons_first_email': self.is_persons_first_email,
        }

        log_args_str = ', '.join([f"{k}={v}" for k, v in log_args.items() if v])
        self._debug_log(f"Truncate determination: {log_args_str}")
        return None if num_chars is None else (0, num_chars)

    @property
    def config(self) -> EmailCfg | None:
        """Configured timestamp, if any."""
        if self.file_info.is_local_extract_file:
            this_file_cfg = cast(EmailCfg, CONFIGS_BY_ID.get(self.file_info.file_id))

            # Merge config for file this was extracted from into self.derived_cfg
            if not self.derived_cfg and (extracted_from_cfg := CONFIGS_BY_ID.get(self.file_info.url_slug)):
                self.derived_cfg = this_file_cfg or EmailCfg(id=self.file_id)

                # Add "appears in [SOURCE]" pfx to description for docs extract ed from others (e.g. huge Dilorio emails)
                if (extracted_from_description := extracted_from_cfg.complete_description):
                    self.derived_cfg.note = f"{APPEARS_IN} {extracted_from_description}"

                for prop in DERIVED_CFG_PROPS_TO_COPY:
                    derived_cfg_val = getattr(self.derived_cfg, prop)

                    if is_bool_prop(prop) and derived_cfg_val is not None:
                        continue  # Don't overwrite booleans

                    extracted_cfg_val = getattr(extracted_from_cfg, prop)
                    setattr(self.derived_cfg, prop, derived_cfg_val or extracted_cfg_val)

                self._log(f"Constructed synthetic config: {self.derived_cfg}")

            return self.derived_cfg or this_file_cfg
        else:
            return super().config

    @property
    def _config(self) -> EmailCfg:
        # TODO: annoying to unnecessarily override superclass just to get python to understanding the types
        return self.config or self.dummy_cfg()

    @property
    def display_text(self) -> str:
        """Config overrides what text should be displayed."""
        return str(self.email_parts)
        return collapse_newlines(self._config.display_text or self.text)

    @property
    def header(self) -> EmailHeader:
        self._header = self._header or self.extract_header()
        return self._header

    @property
    def email_parts(self) -> EmailParts:
        """Separate header chars from the rest of the email text."""
        num_header_lines = self.header.num_header_rows

        if self.header.should_rewrite_header:
            header = self.header.rewrite_header()
            lines = []

            # TODO: Emails w/configured 'actual_text' are particularly broken; need to shuffle some lines
            if (actual_text := self._config.actual_text) is not None:
                lines.extend([cast(str, actual_text), '\n'])
                num_header_lines += 1

            lines.extend(self.lines[num_header_lines:])
            body = _add_line_breaks('\n'.join(lines))
        else:
            header = '\n'.join(self.lines[0:num_header_lines])
            body = '\n'.join(self.lines[num_header_lines:])

        return EmailParts(header, body)

    @property
    def html_margin_bottom(self) -> str:
        """Overloaded in `Email` for case of emails with attachments."""
        if self.attached_docs:
            return '6px'
        else:
            return super().html_margin_bottom

    @property
    def subheaders(self) -> list[Text]:
        """Overloads superclass to avoid returning config.note because that's now in the Panel title."""
        return [self.subheader_info]

    @property
    def is_fwded_article(self) -> bool:
        """True if this email is just a forward of an article from WSJ or whatever."""
        if self._config.is_fwded_article is not None:
            return self._config.is_fwded_article
        elif self._config.fwded_text_after:
            return (len(self.actual_text) < 100)
        else:
            return False

    @property
    def is_interesting(self) -> bool | None:
        """Junk emails are not interesting, configured value takes precedence, fallback to EMAILERS_OF_INTEREST check."""
        if self.is_mailing_list:
            return False
        elif (is_interesting := super().is_interesting) is not None:
            return is_interesting
        elif self.participants.intersection(EMAILERS_OF_INTEREST_SET):
            return True

    @property
    def is_junk_mail(self) -> bool:
        return self.author in JUNK_EMAILERS

    @property
    def is_mailing_list(self) -> bool:
        return self.author in MAILING_LISTS or self.is_junk_mail

    @property
    def is_word_count_worthy(self) -> bool:
        """True if this file should be included in word counts."""
        return not (self.is_fwded_article or self.is_mailing_list or self.is_duplicate)

    @property
    def metadata(self) -> Metadata:
        metadata = {**super().metadata, **self.truthy_props(METADATA_FIELDS)}

        if not self.header.is_empty:
            metadata['header'] = self.header.as_dict()

        if self.attached_docs:
            metadata['attachment_file_ids'] = [f.file_id for f in self.attached_docs]

        return metadata

    @property
    def prettified_txt(self) -> Text:
        """Overrides superclass. Cleaned up / formatted Text ready to be displayed."""
        if self._config.char_range and self._config.char_range[0] > 0:
            if self._config.char_range[0] < self.email_parts.header_len:
                self._warn(f"The excerpt appears to start in the email header which will may in duplicate header chars")

            # always show the email header even if only a configured excerpt is being displayed
            return self.email_parts.header_txt.append('\n\n').append(super().prettified_txt)
        else:
            return super().prettified_txt

    @property
    def recipients_str(self) -> str:
        return ';'.join([str(r) for r in self.recipients])

    @property
    def subheader_info(self) -> Text:
        """String describing this email including author, recipients, and timestamp."""
        author_txt = self.author_txt

        if self._config.author_uncertain:
            author_txt += Text(f" {QUESTION_MARKS}", style=self.author_style)

        prefix = 'fwded article' if self.is_fwded_article else 'email'
        return site_config.email_subheader(prefix, author_txt, self.recipients_txt(), self.timestamp)

    @property
    def subject(self) -> str:
        """String in the Subject: header line."""
        return self._config.subject or self.header.subject or ''

    @property
    def uninteresting_txt(self) -> Text | None:
        """Text to print for uninteresting files."""
        if (uninteresting_txt := super().uninteresting_txt):
            if self.subject:
                uninteresting_txt.append(f' ("{self.subject}")', style='light_yellow3')

            return uninteresting_txt

    @property
    def _summary(self) -> Text:
        """Append the recipients to Document._summary() and close the bracket."""
        txt = self._summary_with_author

        if self.recipients:
            txt.append(', ').append(styled_key_value('recipients', self.recipients_txt()))

        return txt.append(CLOSE_PROPERTIES_CHAR)

    def entity_scan(self, exclude: EntityScanArg = None, include: EntityScanArg = None) -> list[Entity]:
        """Overrides superclass to append names from email attachments."""
        attachment_entities = flatten([ad.entities for ad in self.attached_docs])
        include = Entity.coerce_entity_names(attachment_entities) + Entity.coerce_entity_names(include)
        return super().entity_scan(exclude, include)

    def extract_author(self) -> Name:
        """Overloads superclass method, called at instantiation time."""
        if self.header.author:
            authors = without_falsey(extract_emailer_names(self.header.author))
            return authors[0] if authors else None

    def extract_header(self) -> EmailHeader:
        """Extract an `EmailHeader` from the OCR text."""
        header_match = EMAIL_SIMPLE_HEADER_REGEX.search(self.text)

        if header_match:
            header = EmailHeader.from_header_lines(header_match.group(0))

            # DOJ file OCR text is broken in a less consistent way than the HOUSE_OVERSIGHT files
            if header.is_empty and not self.file_info.is_doj_file:
                header.repair_empty_header(self.lines)
        else:
            self._log_top_lines(msg='No email header found!', level=logging.INFO if self.config else logging.WARNING)
            header = EmailHeader(field_names=[])

        logger.debug(f"{self.file_id} extracted header\n\n{header}\n")
        return header

    def extract_recipients(self) -> list[Name]:
        """Scan the To:, BCC: and CC: fields for known names, falling back to raw strings if no names identified."""
        recipients = uniquify(flatten([extract_emailer_names(r) for r in self.header.recipients]))

        # Assume mailing list emails are to Epstein
        if self.author in BCC_LISTS and (self.is_note_to_self(recipients) or not recipients):
            recipients: list[Name] = [JEFFREY_EPSTEIN]

        # Remove self CCs but preserve self emails
        if self.author and self.author in recipients and not self.is_note_to_self(recipients):
            self._log(f"Removing email to self for {self.author}")
            recipients = [r for r in recipients if r != self.author]

        # Add None to recipients if there's an empty From: or To: header
        if self.header.is_to_redacted and not self.has_unknown_recipient:
            # TODO: SEC is filled in for Dilorio's split up emails after Email is fully instantiated
            if self.author != CHRISTOPHER_DILORIO:
                self._log(f"Appending {None} to recipient list because the To: field is empty")
                recipients.append(None)

        return sort_names(recipients)

    def extract_timestamp(self) -> datetime:
        """Find the time this email was sent by attempting to parse the headers."""
        if self.header.sent_at and (timestamp := _parse_email_timestamp(self.header.sent_at)):
            return timestamp

        searchable_lines = self.lines[0:MAX_NUM_HEADER_LINES]
        searchable_text = '\n'.join(searchable_lines)

        if (date_match := DATE_HEADER_REGEX.search(searchable_text)):
            if (timestamp := _parse_email_timestamp(date_match.group(1))):
                return timestamp

        logger.debug(f"Failed to find timestamp, falling back to parsing {MAX_NUM_HEADER_LINES} lines...")

        for line in searchable_lines:
            if not TIMESTAMP_LINE_REGEX.search(line):
                continue

            if (timestamp := _parse_email_timestamp(line)):
                logger.debug(f"Fell back to timestamp {timestamp} in line '{line}'...")
                return timestamp

        no_timestamp_msg = f"No timestamp found in '{self.file_path.name}'"

        if self._config.duplicate_of_id:
            logger.warning(f"{no_timestamp_msg} but timestamp should be copied from {self._config.duplicate_of_id}")
        else:
            logger.error(f"{no_timestamp_msg}, top lines:\n" + '\n'.join(self.lines[0:MAX_NUM_HEADER_LINES + 10]))

        return FALLBACK_TIMESTAMP

    def file_display(self, align: JustifyMethod | None = None) -> FileDisplay:
        """Allows for proper right vs. left justify."""
        return FileDisplay(
            body_panel=self._body_as_table(),
            document=self,
            file_info=self.file_id_panel,
            indent=site_config.indents.info,
            justify=align,
            margin_bottom=self.html_margin_bottom,
            subheaders=self.subheaders,
        )

    def is_from_or_to(self, name: str) -> bool:
        """True if `name` is either the author or one of the recipients."""
        return name in self.participants

    def is_note_to_self(self, recipients: list[Name] | None = None) -> bool:
        return (recipients if recipients is not None else self.recipients) == [self.author]

    def recipients_txt(self, max_full_names: int = 2) -> Text:
        """Comma separated colored names (last name only if more than `max_full_names` recipients)."""
        recipients = [r or UNKNOWN for r in self.recipients] if len(self.recipients) > 0 else [UNKNOWN]

        names = [
            Text(r if len(recipients) <= max_full_names else extract_last_name(r), get_style_for_name(r)) + \
                (Text(f" {QUESTION_MARKS}", get_style_for_name(r)) if self._config.recipient_uncertain else Text(''))
            for r in recipients
        ]

        return join_texts(names, join=', ')

    def to_html(self) -> str:
        html = super().to_html()

        if (attachments_table := self._attached_docs_table()):
            indent_props = {'margin-left': to_em(site_config.indents.email_attachment)}
            html += table_to_html(attachments_table, indent_props)

        return html

    def _attached_docs_table(self) -> Table | None:
        if not self.attached_docs:
            return None

        attachments_table_title = f"| Email Attachments for {self.file_info.url_slug}:" # ╏┇┣

        if (doc := self.attached_docs[0])._config.show_full_panel:
            if len(self.attached_docs) > 1:
                raise ValueError(f"Can't show more than one panelized attachment for {self}!")

            table = Table(title=attachments_table_title, title_justify='left')
            table.add_column(doc._config.note)
            table.add_row(highlighter(Text(self.attached_docs[0].text, EXCERPT_STYLE)))
            return table
        else:
            return OtherFile.files_preview_table(self.attached_docs, title=attachments_table_title)

    def _body_as_table(self) -> Table:
        """Renders the info text as a top row in a table-ish view."""
        note_txt = Text('', justify='right').append(self._config.note_txt) if self._config.note_txt else ''

        panel = Table(
            border_style=self.border_style,
            box=box.ROUNDED,
            header_style='on gray11',
            show_header=bool(note_txt),
        )

        panel.add_column(note_txt)
        panel.add_row(self.prettified_txt)
        return panel

    def _debug_props(self) -> DebugDict:
        props = super()._debug_props()
        local_props = self.truthy_props(DEBUG_PROPS)
        local_props['extracted_recipients'] = self.extracted_recipients

        if not self.header.is_empty:
            local_props['header'] = self.header.as_dict()
        if self.is_note_to_self():
            local_props['is_note_to_self'] = self.is_note_to_self()

        props.update(prefix_keys(self._debug_prefix, local_props))
        return props

    def _extract_actual_text(self) -> str:
        """The text that comes before likely quoted replies and forwards etc."""
        if self._config.actual_text is not None:
            return self._config.actual_text

        text = '\n'.join(self.text.split('\n')[self.header.num_header_rows:]).strip()

        if self._config.fwded_text_after:
            return text.split(self._config.fwded_text_after)[0].strip()
        elif self.header.num_header_rows == 0:
            return self.text

        self._log_top_lines(20, "Raw text:", logging.DEBUG)
        self._log(f"With {self.header.num_header_rows} header lines removed:\n{text[0:500]}\n\n", logging.DEBUG)
        reply_text_match = REPLY_TEXT_REGEX.search(text)

        if reply_text_match:
            actual_num_chars = len(reply_text_match.group(1))
            actual_text_pct = f"{(100 * float(actual_num_chars) / len(text)):.1f}%"
            logger.debug(f"'{self.file_id}': actual_text() reply_text_match is {actual_num_chars:,} chars ({actual_text_pct} of {len(text):,})")
            text = reply_text_match.group(1)

        # If all else fails look for lines like 'From: blah', 'Subject: blah', and split on that.
        for field_name in REPLY_SPLITTERS:
            field_string = f'\n{field_name}'

            if field_string not in text:
                continue

            pre_from_text = text.split(field_string)[0]
            actual_num_chars = len(pre_from_text)
            actual_text_pct = f"{(100 * float(actual_num_chars) / len(text)):.1f}%"
            logger.debug(f"'{self.file_id}': actual_text() fwd_text_match is {actual_num_chars:,} chars ({actual_text_pct} of {len(text):,})")
            text = pre_from_text
            break

        return text.strip()

    def _idx_of_nth_quoted_reply(self, n: int = MAX_QUOTED_REPLIES) -> int | None:
        """Get position of the nth 'On June 12th, 1985 [SOMEONE] wrote:' style line in self.text."""
        header_offset = len(self.header.header_chars)
        text = self.text[header_offset:]

        for i, match in enumerate(QUOTED_REPLY_LINE_REGEX.finditer(text)):
            if i >= n:
                return match.end() + header_offset - 1

    def _remove_bad_lines(self) -> None:
        """Get rid of crufty lines matching `BAD_LINE_REGEX`"""
        self._set_text(lines=[line for line in self.lines if not BAD_LINE_REGEX.match(line)])

    def _repair(self) -> None:
        """Repair particularly janky files. Note that OCR_REPAIRS are applied *after* other line adjustments."""
        # for line in self.lines[0:NUM_LINES_TO_REPAIR_HEADERS]:
        #     if HEADER_REPAIR_REGEX.match(line):
        #         self._warn(f'reparable header line: {quote(line)}')

        super()._repair()

        # Apply custom repairs for DOJ files
        if self.file_info.is_doj_file:
            new_text = DojFile._remove_bad_lines(strip_pdfalyzer_panels(self.text))
            self._set_text(text=self.repair_ocr_text(DOJ_EMAIL_OCR_REPAIRS, new_text))

        if BAD_FIRST_LINE_REGEX.match(self.lines[0]):
            self._set_text(lines=self.lines[1:])

        self._remove_bad_lines()
        self.__bespoke_repair_house_oversight_emails()
        self._set_text(text=self.repair_ocr_text(OCR_REPAIRS, self.text))

        if self.file_id in SINGLE_EMAIL_OCR_REPAIRS:
            self._debug_log('applying bespoke OCR repairs')
            self._set_text(text=self.repair_ocr_text(SINGLE_EMAIL_OCR_REPAIRS[self.file_id], self.text))

        self._repair_links_and_quoted_subjects()
        self._format_newlines_and_snip_signatures()

    def _repair_links_and_quoted_subjects(self) -> None:
        """Repair links that the OCR has broken into multiple lines as well as 'Subject:' lines."""
        subject_line = next((line for line in self.lines if line.startswith('Subject:')), None) or ''
        subject = subject_line.split(':')[1].strip() if subject_line else ''
        lines = self.lines
        new_lines = []
        i = 0

        while i < len(lines):
            line = lines[i]

            if LINK_LINE_REGEX.search(line):
                while i < (len(lines) - 1) \
                        and not (lines[i + 1].startswith('htt') or lines[i + 1].startswith('www') or lines[i + 1].endswith('com')) \
                        and (lines[i + 1].endswith('/') \
                            or any(s in lines[i + 1] for s in URL_SIGNIFIERS) \
                            or LINK_LINE2_REGEX.match(lines[i + 1])):
                    logger.debug(f"{self.filename}: Joining link lines\n   1. {line}\n   2. {lines[i + 1]}\n")
                    line += lines[i + 1]
                    i += 1

                line = line if '[' in line or ']' in line else line.replace(' ', '')
            elif ' http' in line and line.endswith('html'):
                pre_link, post_link = line.split(' http', 1)
                line = f"{pre_link} http{post_link.replace(' ', '')}"
            elif line.startswith('Subject:') and i < (len(lines) - 2) and len(line) >= 40:
                next_line = lines[i + 1]
                next_next = lines[i + 2]

                if len(next_line) <= 1 or any([cont in next_line for cont in BAD_SUBJECT_CONTINUATIONS]):
                    pass
                elif (subject.endswith(next_line) and next_line != subject) \
                        or (HEADER_FIELD_COLON_REGEX.search(next_next) and not HEADER_FIELD_COLON_REGEX.search(next_line)):
                    self._log(f"Fixing broken subject line\n  line: '{line}'\n    next: '{next_line}'\n    next: '{next_next}'\nsubject='{subject}'\n")
                    line += f" {next_line}"
                    i += 1

            new_lines.append(line)
            i += 1

        self._debug_log(f"----after line repair---\n" + '\n'.join(new_lines[0:20]) + "\n---")
        self._set_text(lines=new_lines)

    def _sent_from_device(self) -> str | None:
        """Find any 'Sent from my iPhone' style signature line if it exist in the 'actual_text'."""
        if (sent_from_match := SENT_FROM_REGEX.search(self.actual_text)):
            sent_from = sent_from_match.group(0).strip()

            if sent_from.startswith('Typos, misspellings courtesy of iPhone'):  # special handling for linda stone periods
                return sent_from.removesuffix('.')
            elif sent_from.startswith('sent'):
                return capitalize_first(sent_from)
            else:
                return sent_from

    def _format_newlines_and_snip_signatures(self) -> None:
        """Add newlines before quoted replies, snip signatures and XML, etc.."""
        # Insert line breaks now unless header is broken, in which case we'll do it later after fixing header
        # self._debug_log(f"text before _add_line_breaks:\n\n{self.text}\n---")
        text = self.text if self.header.was_initially_empty else _add_line_breaks(self.text)
        text = REPLY_REGEX.sub(r'\n\1', text)  # Newlines between quoted replies
        text = FORWARDED_TOO_MUCH_SPACE_REGEX.sub(r'\1\n', text)
        # self._debug_log(f"text after _add_line_breaks:\n\n{text}\n---")

        for name, signature_regex in EMAIL_SIGNATURE_REGEXES.items():
            signature_replacement = f'<...snipped {name.lower()} email signature...>'
            text, num_replaced = signature_regex.subn(signature_replacement, text)
            self.signature_substitution_counts[name] = self.signature_substitution_counts.get(name, 0)
            self.signature_substitution_counts[name] += num_replaced

        # Remove XML cruft in some files
        text, num_plists_stripped = XML_PLIST_REGEX.subn(XML_STRIPPED_MSG, text)

        if num_plists_stripped:
            self._log(f"Replaced {num_plists_stripped} XML plists...")

        self._set_text(text=collapse_newlines(text).strip())
        self._remove_bad_lines()  # TODO: we're currently calling this twice to handle lines with HTML garbage that turn into something matching BAD_LINE_REGEX

    def __bespoke_repair_house_oversight_emails(self) -> None:
        """Apply destructive repairs to the underyling text programmtically because we don't want to edit the underlying file."""
        old_text = self.text

        if self.file_id in LINE_REPAIR_MERGES:
            for merge_args in LINE_REPAIR_MERGES[self.file_id]:
                self.__merge_lines(*merge_args)

        if self.file_id in ['025233']:
            self.lines[4] = f"Attachments: {self.lines[4]}"
            self._set_text(lines=self.lines)
        elif self.file_id == '029977':
            self._set_text(text=self.text.replace('Sent 9/28/2012 2:41:02 PM', 'Sent: 9/28/2012 2:41:02 PM'))
        elif self.file_id == '025041':
            self.__remove_line(4)
            self.__remove_line(4)
        elif self.file_id == '029692':
            self.__remove_line(3)

        if old_text != self.text:
            self._log(f"Modified text, old:\n\n" + '\n'.join(old_text.split('\n')[0:12]) + '\n')
            self._log_top_lines(12, 'Result of modifications')

    def __merge_lines(self, idx1: int, idx2: int | None = None) -> None:
        """Combine lines numbered 'idx' and 'idx2' into a single line (idx2 defaults to idx + 1)."""
        if idx2 is None:
            self._line_merge_arguments.append((idx1,))
            idx2 = idx1 + 1
        else:
            self._line_merge_arguments.append((idx1, idx2))

        if idx2 < idx1:
            lines = self.lines[0:idx2] + self.lines[idx2 + 1:idx1] + [self.lines[idx1] + ' ' + self.lines[idx2]] + self.lines[idx1 + 1:]
        elif idx2 == idx1:
            raise RuntimeError(f"idx2 ({idx2}) must be greater or less than idx ({idx1})")
        else:
            lines = self.lines[0:idx1]

            if idx2 == (idx1 + 1):
                lines += [self.lines[idx1] + ' ' + self.lines[idx1 + 1]] + self.lines[idx1 + 2:]
            else:
                lines += [self.lines[idx1] + ' ' + self.lines[idx2]] + self.lines[idx1 + 1:idx2] + self.lines[idx2 + 1:]

        self._set_text(lines=lines)

    def __remove_line(self, idx: int) -> None:
        """Remove a line from `self.lines`."""
        num_lines = idx * 2
        self._log_top_lines(num_lines, msg=f'before removal of line {idx}')
        del self.lines[idx]
        self._set_text(lines=self.lines)
        self._log_top_lines(num_lines, msg=f'after removal of line {idx}')

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        yield self.rich_header()
        body_bottom_padding = 0 if self.attached_docs else 1
        yield Padding(self._body_as_table(), (0, 0, body_bottom_padding, site_config.indents.other_files_table))

        if (attachments_table := self._attached_docs_table()):
            yield Padding(attachments_table, (0, 0, 1, site_config.indents.email_attachment))

    @classmethod
    def dummy_cfg(cls) -> EmailCfg:
        return EmailCfg(id='DUMMY')

    @staticmethod
    def build_emails_table(emails: list['Email'], name: Name = '', title: str = '', show_length: bool = False, **kwargs) -> Table:
        """Turn a list of `Email` objects into a `Table` with sender, recipient, and subject line."""
        if title and name:
            raise ValueError(f"Can't provide both 'author' and 'title' args")
        elif name == '' and title == '':
            raise ValueError(f"Must provide either 'author' or 'title' arg")

        author_style = get_style_for_name(name, allow_bold=False)
        link_style = author_style if name else ARCHIVE_LINK_COLOR
        min_width = len(name or UNKNOWN)
        max_width = max(20, min_width)

        columns = [
            {'name': 'Sent At', 'justify': 'left', 'style': TIMESTAMP_DIM},
            {'name': 'From', 'justify': 'left', 'min_width': min_width, 'max_width': max_width},
            {'name': 'To', 'justify': 'left', 'min_width': min_width, 'max_width': max_width + 2},
            {'name': 'Length', 'justify': 'right', 'style': 'wheat4'},
            {'name': 'Subject', 'justify': 'left', 'min_width': 35, 'style': 'honeydew2'},
        ]

        table = build_table(
            title or None,
            cols=[col for col in columns if show_length or col['name'] not in ['Length']],
            border_style=DEFAULT_TABLE_KWARGS['border_style'] if title else author_style,
            header_style="bold",
            highlight=True,
            **kwargs
        )

        for email in emails:
            fields = [
                link_text_obj(email.file_info.external_url, email.timestamp_without_seconds, style=link_style),
                email.author_txt,
                email.recipients_txt(max_full_names=1),
                f"{email.length}",
                email.subject,
            ]

            if not show_length:
                del fields[3]

            table.add_row(*fields)

        return table


def _add_line_breaks(email_text: str) -> str:
    text = EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX.sub(r'\n\1\n', email_text).strip()
    # logger.debug(f"text after EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX _add_line_breaks()\n---\n{text}\n---")
    return FORWARDED_TOO_MUCH_SPACE_REGEX.sub(r'\1\n', text)


# TODO: this might be obsolete bc of usage of datutil or similar
def _parse_email_timestamp(timestamp_str: str) -> None | datetime:
    try:
        if (american_date_match := AMERICAN_TIME_REGEX.search(timestamp_str)):
            timestamp_str = american_date_match.group(1)
        else:
            timestamp_str = timestamp_str.replace('(GMT-05:00)', 'EST')
            timestamp_str = BAD_TIMEZONE_REGEX.sub(' ', timestamp_str).strip()

        timestamp = parse(timestamp_str, fuzzy=True, tzinfos=TIMEZONE_INFO)
        logger.debug(f'Parsed timestamp "%s" from string "%s"', timestamp, timestamp_str)
        return coerce_utc(timestamp)
    except Exception as e:
        logger.debug(f'Failed to parse "{timestamp_str}" to timestamp!')
        return None



from epstein_files.documents.config.config_builder import fbi_defense_witness, fbi_interview, fbi_tip, fbi_report, letter
from epstein_files.documents.config.doc_cfg import NO_TRUNCATE, SHORT_TRUNCATE_TO, DocCfg
from epstein_files.documents.config.email_cfg import EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import MINOR_VICTIM, REDACTED
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.util.logging import logger

FBI_REPORT_FIELDS = [
    'Case ID #',
    'Drafted By',
    'Date/Time Received',
    'Details',
    'SentinelCaseld',
    'SentToSentinel',
]


def bop_doc(id: str, description: str = '', date: str = '', display_text: str = '', **kwargs) -> DocCfg:
    return DocCfg(
        id=id,
        author=BUREAU_OF_PRISONS,
        note=description,
        date=date,
        display_text=display_text,
        **kwargs
    )


def bop_policy_doc(id: str, display_text: str, date: str = '') -> DocCfg:
    """Bureau of prison brochures, poliicy statements, etc."""
    return bop_doc(id, date=date, display_text=display_text)


GOVERNMENT_CFGS = [
    DocCfg(
        id='024117',
        note=f"anti-money laundering (AML), Bank Secrecy Act (BSA), & terrorist financing (CFT) US law FAQ",
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA00315076',
        date='2008-06-01',
        date_uncertain=True,
        note=f"visitor list during Epstein's 2009 incarceration in {PALM_BEACH}",
        show_full_panel=True,
    ),
    DocCfg(
        id='EFTA00306074',
        date='2008-06-01',
        date_uncertain=True,
        note=f"approved mail list during Epstein's 2009 incarceration in {PALM_BEACH}",
        show_full_panel=True,
    ),
    bop_doc('EFTA00109783', 'prisoner assignments', '2019-08-03'),
    bop_doc('EFTA00035921', "Lieutenant's Logs", '2019-08-06'),
    bop_doc('EFTA00039153', 'List of Exhibits, Chapter 2', '2019-01-06'),
    bop_doc('EFTA00109163', 'Metropolitan Correctional Center forms showing Konstantin Ignatov', '2019-08-08', is_interesting=True),
    bop_doc('EFTA00120617', '"Prisoner Remand or Order to Deliver" forms', '2019-07-30'),
    bop_doc('EFTA00109654', 'roster of inmates at Metropolitan Correctional Center', '2019-08-08'),
    bop_doc('EFTA00108533', 'roster of inmates at Metropolitan Correctional Center', '2019-07-23'),
    bop_doc('EFTA00108552', 'roster of inmates at Metropolitan Correctional Center', '2019-07-23'),
    bop_doc('EFTA00039025', "report on death of Jeffrey Epstein", '2023-06-22', is_interesting=10),
    bop_doc(
        'EFTA00034275',
        "initial form about Epstein's suicide attempt",
        '2019-07-23',
        comment='visible in EFTA00019348',
        is_interesting=False,
    ),
    bop_doc(
        'EFTA00019348',
        "report on Epstein's failed suicide attempt",
        '2019-08-02',
        highlight_quote='staff observed inmate Epstein, Jeffrey lying in the fetal position on the floor with a homemade fashioned noose around his neck',
        truncate_to=(7_500, 12_500),
    ),
    bop_doc('EFTA00039190', 'Special Housing Units', '2016-11-23', is_interesting=False),
    bop_policy_doc('EFTA00120459', 'handwritten log of prisoner movements', date='2019-08-09'),
    bop_policy_doc('EFTA00039227', 'Inmate Discipline Program Statement'),
    bop_policy_doc('EFTA00039295', 'Inmate Telephone Privileges Program Statement'),
    bop_policy_doc('EFTA00039312', 'Program Statement / Memo about BOP Pharmacy Program'),
    bop_policy_doc('EFTA00039351', 'Program Statement / Memo about BOP Pharmacy Program', date='2004-11-17'),
    bop_policy_doc('EFTA00039156', 'Standards of Employee Conduct'),
    DocCfg(
        id='EFTA01125108',
        author='DHS',
        note=f'receipt for I129 Petition for a Nonimmigrant Worker filed by Sublime Art LLC, {ARDA_BESKARDES}',
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA00164939',
        author=DOJ,
        date='2025-09-01',
        date_uncertain='approximate',
        note='Powerpoint summary of Child Sex Trafficking Task Force Epstein investigation',
        is_interesting=True,
    ),
    DocCfg(id='EFTA02731200', author=DOJ, note="memo about potential prosecution of Epstein's assistant"),
    DocCfg(id='EFTA02731082', author=DOJ, note="memo about investigation into Epstein's co-conspirators"),
    DocCfg(id='EFTA02730741', author=DOJ, date='2025-05-01', date_uncertain=True, note="Evidence list for 50D-NY-3027571 Filtering On 'Type(s): 1B'"),
    DocCfg(id='EFTA02730486', author=DOJ, date='2025-05-01', date_uncertain=True, note="Evidence list for 50D-NY-3027571 Filtering On '1A'"),
    DocCfg(id='EFTA00040006', author=DOJ, date='2019-08-27', note='Personal History of Defendant Jeffrey Epstein + grand jury indictment'),
    DocCfg(id='EFTA02731226', author=DOJ, date='2021-03-14', note=f'memo seeking authorization to charge {GHISLAINE_MAXWELL} with additional offenses'),
    DocCfg(id='EFTA00023055', author=FBI, note="evidence of notes left about newly recruited underage girls by girls giving massages"),
    DocCfg(id='EFTA01731217', author=FBI, note=f'requesting INS allow {NADIA_MARCINKO} be allowed to stay in the US because of an ongoing sex-trafficking case', is_interesting=True),
    DocCfg(id='EFTA00247131', author=FBI, note='search warrant for New York house', date='2019-07-07'),
    fbi_defense_witness('EFTA02730267', 'Malcolm Grumbridge', '2022-04-14'),
    fbi_defense_witness('EFTA02730477', 'Roderic Alexander', '2022-01-19'),
    fbi_defense_witness('EFTA02730271', None, '2022-03-22'),
    # FBI interviews
    fbi_interview('EFTA02858481', 'Jennifer Aros', 'Epstein victim'),
    fbi_interview('EFTA00174375', LUKE_D_THORBURN, f"lots of takes on Epstein, China, and {STEVE_BANNON}"),
    fbi_interview('EFTA00081226', MINOR_VICTIM),
    fbi_interview('EFTA00038915', MINOR_VICTIM, 'claims Epstein knew she was 14'),
    fbi_interview(
        'EFTA01246710',
        PERRY_LANG,
        "Epstein's chef claims Donald Trump came to Epstein's house for dinner",
        truncate_to=(6_000, 7_500),
    ),
    fbi_interview('EFTA00101927', None, f"claims Glenn and {EVA_DUBIN}'s Swiss au pair was being held against her will"),
    fbi_interview('EFTA00159321', None, f'covers {PAOLO_ZAMPOLLI}, Epstein, and the possibility Epstein introduced Melania to Donald Trump'),
    # FBI reports
    fbi_report(
        '018872',
        non_participants=[
            BILL_GATES,
            BILL_RICHARDSON,
            EDUARDO_ROBLES,
            'Eliot Spitzer',
            GERALD_LEFCOURT,
            GLENN_DUBIN,
            JEAN_LUC_BRUNEL,
            JOI_ITO,
            LARRY_SUMMERS,
            LAWRANCE_VISOSKI,  # Just bc his deposition comes soon after
            MARK_EPSTEIN,
            MARTIN_NOWAK,
            MORTIMER_ZUCKERMAN,
            PRINCE_ANDREW,
            SECURITIES_AND_EXCHANGE_COMMISSION,
            STEVEN_HOFFENBERG,
            SVETLANA_POZHIDAEVA
        ],
    ),
    fbi_report('021569'),
    fbi_report('021434', is_valid_for_name_scan=False),
    fbi_report('019352', f"contains clippings of various press items about Epstein"),
    fbi_report(
        'EFTA00129085',
        'wiretap linking phone number in John Gotti / Gambino / Michael Bilotti investigation to phone in Epstein investigation',
        is_interesting=True,
    ),
    fbi_report('EFTA01688746'),
    fbi_report('EFTA00151754', 'declaration of Law Enforcement Officer for Victim of Trafficking', is_interesting=True),
    fbi_report('EFTA00222943', "agent believes computers were removed from Epstein's residence"),
    fbi_report('EFTA01249591', f"allegations against {HENRY_JARECKI}", show_full_panel=True),
    fbi_report('EFTA00173569', 'hack of FBI Epstein files repository by foreign actor', is_interesting=True),
    fbi_report('EFTA00020506', highlight_quote='chauffeur also told Epstein "I have something on you remember what I buried!"'),
    # FBI tips
    fbi_tip(
        'EFTA01660676',
        "about recently convicted rapists Tal and Oren Alexander at Epstein's house",
        show_full_panel=True,
        url='https://www.bbc.com/news/articles/c6271ngl014o',
    ),
    fbi_tip('EFTA00020490', 'from woman who thinks she encountered Epstein as a young girl'),
    fbi_tip('EFTA00090314', f'about {MASHA_DROKOVA}, Jared Kushner, Ivanka Trump, Chabad, {ALAN_DERSHOWITZ}, etc.', is_interesting=True),
    DocCfg(id='EFTA00084614', author=PALM_BEACH_POLICE, note='incident report detailing the investigation into Jeffrey Epstein'),
    DocCfg(id='EFTA00007893', author=PALM_BEACH_POLICE, note=f"receipts, notes, bank statements of {GHISLAINE_MAXWELL}"),
    DocCfg(id='EFTA00005569', author=PALM_BEACH_POLICE, display_text='photo lineup featuring Epstein', date='2005-03-17'),
    DocCfg(id='EFTA00003868', author=PALM_BEACH_POLICE, display_text='photo lineup featuring Epstein', date='2006-08-09'),
    DocCfg(id='EFTA01227877', note='multi entry visa for the Russian Federation', date='2018-06-25', show_full_panel=True),
    DocCfg(id='EFTA00128379', note='analysis of two of Epstein\'s desktop computers', date='2019-01-06', is_interesting=True),
    DocCfg(id='EFTA02730274', note='evidence inventory that appears to have since been deleted from the DOJ website'),
    DocCfg(id='EFTA00263284', note='notes about deceit and sexual manipulation by Australian professor Vincent Bulone', is_interesting=True),
    DocCfg(id='EFTA00001884', note='photo of letter from Virgin Islands DOJ to Epstein', date='2019-03-14'),
    DocCfg(id='EFTA00074744', note="USVI court filing about Epstein will and estate"),
    DocCfg(id='EFTA00007157', note='victim list and police log'),
    letter('EFTA01653121', FBI, ['USCIS'], "regarding an individual's cooperation in the investigation of Epstein and Maxwell"),
    letter('EFTA00098456', PAUL_G_CASSELL, ['Scotland Yard'], 'International Sex Trafficking by Jeffrey Epstein, contains court filings'),

    # Emails
    EmailCfg(
        id='EFTA00129096',
        date='2025-04-03 7:13:35 PM',
        note=f'background check on Tim Draper and {MASHA_DROKOVA}',
        show_full_panel=True,
        truncate_to=NO_TRUNCATE,  # TODO this shouldn't be necessary?
    ),
    EmailCfg(id='EFTA02730483', author=FBI, date='2023-07-11T08:25:00', date_uncertain='actually reply timestamp'),
    EmailCfg(id='EFTA02731552', author=FBI, recipients=[USANYS], date='2021-05-26 16:12:00', recipient_uncertain=True),
    EmailCfg(id='EFTA00039971', author=FBI, recipients=[USANYS], recipient_uncertain=True),
    EmailCfg(id='EFTA00037683', note=f"tip that the murder of DC Madam Jeanne Palfrey might be connected to Epstein's network"),

    # DOJ / USANYS emails
    EmailCfg(id='EFTA00039967', author='DOJ London', recipients=[USANYS]),
    EmailCfg(
        id='EFTA00039660',
        author='DOJ Chief Psychologist',
        note="report on Epstein's psychological state in jail",
        is_interesting=True,
        recipients=[USANYS],
    ),
    EmailCfg(id='EFTA00040145', date='2021-11-09 17:24:30', is_interesting=False),
    EmailCfg(
        id='EFTA02731689',
        author=USANYS,
        date='2023-06-09 20:14:00',
        author_uncertain=True,
        recipients=[USANYS],
        truncate_to=SHORT_TRUNCATE_TO,
    ),
    EmailCfg(
        id='EFTA00277089',
        author=USANYS,
        recipients=[USANYS],
        note="tip for the FBI / DOJ that Swedish billionaire Johan Eliasch may be connected to Epstein's crimes",
        duplicate_ids=['EFTA00151184']
    ),
    EmailCfg(
        id='EFTA00157083',
        author=USANYS,
        highlight_quote='STXU has recovered a password "asdfasdfasdfasdf" (no quotes) for item NYCO24364',
        truncate_to=1400,
    ),
    EmailCfg(id='EFTA02731755', author=USANYS),
    EmailCfg(id='EFTA00039662', author=USANYS),
    EmailCfg(id='EFTA00040141', author=USANYS),
    EmailCfg(id='EFTA02731644', author=USANYS),
    EmailCfg(id='EFTA00040144', author=USANYS),
    EmailCfg(id='EFTA00039995', author=USANYS),
    EmailCfg(id='EFTA00039813', author=USANYS),
    EmailCfg(id='EFTA00039890', author=USANYS, is_interesting=False),
    EmailCfg(id='EFTA00039815', author=USANYS, is_interesting=False),
    EmailCfg(id='EFTA00039825', author=USANYS, is_interesting=False),
    EmailCfg(id='EFTA02731783', author=USANYS, date='2022-01-21 17:28:00'),
    EmailCfg(id='EFTA02731578', author=USANYS, date='2021-05-28 10:00:00'),
    EmailCfg(id='EFTA02730468', author=USANYS, recipients=[USANYS], date='2019-07-11T08:25:00', date_uncertain='just wrong'),
    EmailCfg(id='EFTA00039983', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA00039886', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731651', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731775', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731646', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731640', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731643', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731501', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA02731633', author=USANYS, author_uncertain=True),
    EmailCfg(id='EFTA01660679', author=USANYS, author_uncertain=True, recipients=[FBI], recipient_uncertain=True),
    EmailCfg(id='EFTA02731699', author=USANYS, author_uncertain=True, recipients=[FBI], date='2021-05-27 10:19:00'),
    EmailCfg(id='EFTA00040121', author=USANYS, recipients=[ATT_COURT_APPEARANCE_TEAM]),
    EmailCfg(id='EFTA02731630', author=USANYS, recipients=[FBI]),
    EmailCfg(
        id='EFTA02731737',
        author=USANYS,
        author_uncertain=True,
        date='2023-06-30T16:05:00',
        is_interesting=True,
        recipients=[FBI],
        truncate_to=5000,
    ),
    EmailCfg(
        id='EFTA00090614',
        author=USANYS,
        author_uncertain=True,
        note='alleges cover up of crimes of wealthy / powerful people by DA Cyrus Vance',
        recipients=[FBI],
        recipient_uncertain=True,
        truncate_to=7000,
    ),
    EmailCfg(id='EFTA02731593', author=USANYS, recipients=['Manhattan DA']),
    EmailCfg(id='EFTA00039419', author=USANYS, recipients=['Manhattan DA']),
    EmailCfg(id='EFTA02731617', author=USANYS, recipients=[SDNY], date='2021-04-28T15:05:41'),
    EmailCfg(id='EFTA00151184', author=USANYS, recipients=[USANYS]),
    EmailCfg(id='EFTA02731615', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731684', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731485', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731736', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731478', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA00039872', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA00039868', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731490', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731477', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731660', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA00039820', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731638', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731636', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731637', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731659', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731781', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731774', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731724', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731771', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731486', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731604', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731618', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(
        id='EFTA02731655',
        author=USANYS,
        author_uncertain=True,
        note=f"specific allegations against {LEON_BLACK}",
        is_interesting=True,
        recipients=[USANYS],
    ),
    EmailCfg(id='EFTA02731608', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731612', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731729', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731484', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731757', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA02731623', author=USANYS, recipients=[USANYS], author_uncertain=True),
    EmailCfg(id='EFTA00039879', author=USANYS, recipients=[USANYS], recipient_uncertain=True),
    EmailCfg(
        id='EFTA02731488',
        author=USANYS,
        author_uncertain=True,
        note=f"possible evidence of girls being trafficked to {LEON_BLACK}",
        recipients=[USANYS],
        is_interesting=True,
    ),
    EmailCfg(id='EFTA02731733', author=USANYS, recipients=[USANYS], date='2021-05-17 17:29:00', author_uncertain=True),
    EmailCfg(id='EFTA02731754', author=USANYS, recipients=[USANYS], date='2024-03-06T23:24:00', author_uncertain=True),
    EmailCfg(id='EFTA02731732', author=USANYS, recipients=[USANYS], date='2024-03-06T12:21:00', author_uncertain=True),
    EmailCfg(id='EFTA02731528', author=USANYS, recipients=[USANYS], date='2021-05-06 09:39:15', author_uncertain=True),
    EmailCfg(id='EFTA02731475', author=USANYS, recipients=[USANYS], date='2023-05-31T20:53:00', author_uncertain=True),
    EmailCfg(id='EFTA02731697', author=USANYS, recipients=[USANYS], date='2021-06-07 17:33:00', author_uncertain=True),
    EmailCfg(id='EFTA02731583', author=USANYS, recipients=[USANYS], date='2022-01-21 17:28:00', author_uncertain=True),
    EmailCfg(id='EFTA02731587', author=USANYS, recipients=[USANYS], date='2022-01-21 17:28:00', author_uncertain=True),
]



from pathlib import Path

from rich.text import Text

from epstein_files.documents.document import Document
from epstein_files.output.rich import console
from epstein_files.util.constant.strings import HOUSE_OVERSIGHT_PREFIX
from epstein_files.util.env import DOCS_DIR
from epstein_files.util.helpers.file_helper import diff_files, extract_file_id


def diff_documents(files: list[str]) -> None:
    """Diff the contents of two `Document` objects after all cleanup, BOM removal, etc."""
    if len(files) != 2:
        raise RuntimeError('Need 2 files')
    elif files[0] == files[1]:
        raise RuntimeError(f"Filenames are the same!")

    files = [f"{HOUSE_OVERSIGHT_PREFIX}{f}" if len(f) == 6 else f for f in files]
    files = [f if f.endswith('.txt') else f"{f}.txt" for f in files]
    tmpfiles = [Path(f"tmp_{f}") for f in files]
    docs = [Document(DOCS_DIR.joinpath(f)) for f in files]

    # Write temporary plain text file with all repairs applied.
    for i, doc in enumerate(docs):
        doc._write_clean_text(tmpfiles[i])

    diff_files(tmpfiles[0], tmpfiles[1])
    console.print(f"Possible suppression with: ")
    console.print(Text('   suppress left: ').append(f"   '{extract_file_id(files[0])}': 'the same as {extract_file_id(files[1])}',", style='cyan'))
    console.print(Text('  suppress right: ').append(f"   '{extract_file_id(files[1])}': 'the same as {extract_file_id(files[0])}',", style='cyan'))

    for f in tmpfiles:
        f.unlink()
